/*
 *          mxpcdrv.c  -- MOXA PC104 Communication Module Linux driver.
 *
 *      Copyright (C) 1999-2007  Moxa Technologies (support@moxa.com.tw).
 *
 *      This code is loosely based on the Linux serial driver, written by
 *      Linus Torvalds, Theodore T'so and others.
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	Original release	03/20/03
 *
 *	03/20/03	Support MOXA PC104 Communication Module.
 *			
 */

#ifdef 		MODVERSIONS
#ifndef 	MODULE
#define 	MODULE
#endif
#endif

#include <linux/version.h>
#define	VERSION_CODE(ver,rel,seq)	((ver << 16) | (rel << 8) | seq)

#ifdef MODULE
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,18))
#include <linux/config.h>
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define	MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
#include <linux/autoconf.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#include "mxpcdrv.h"
#define	MXPCDRV_VERSION	"1.2"
#define	MXPCDRVMAJOR	 34
#define	MXPCDRVCUMAJOR	 41	

#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
#define copy_from_user memcpy_fromfs
#define copy_to_user memcpy_tofs
#define put_to_user(arg1, arg2) put_fs_long(arg1, (unsigned long *)arg2)
#define get_from_user(arg1, arg2) arg1 = get_fs_long((unsigned long *)arg2)
#define schedule_timeout(x) {current->timeout = jiffies + (x); schedule();}
#define signal_pending(x) ((x)->signal & ~(x)->blocked)
#else
#include <asm/uaccess.h>
#define put_to_user(arg1, arg2) put_user(arg1, (unsigned long *)arg2)
#define get_from_user(arg1, arg2) get_user(arg1, (unsigned int *)arg2)
#endif

#define	MXPCDRV_EVENT_TXLOW	 1
#define	MXPCDRV_EVENT_HANGUP	 2

#define SERIAL_DO_RESTART

#define MXPCDRV_BOARDS		4	/* Max. boards */
#define MXPCDRV_PORTS		32	/* Max. ports */
#define MXPCDRV_PORTS_PER_BOARD	8	/* Max. ports per board*/
#define MXPCDRV_ISR_PASS_LIMIT	99999L
#define MXPCDRV_BOARD_INDEX			4

#define	MXPCDRV_ERR_IOADDR	-1
#define	MXPCDRV_ERR_IRQ		-2
#define	MXPCDRV_ERR_IRQ_CONFLIT	-3
#define	MXPCDRV_ERR_VECTOR	-4

#define SERIAL_TYPE_NORMAL	1
#define SERIAL_TYPE_CALLOUT	2

#define WAKEUP_CHARS		256

#define UART_MCR_AFE		0x20
#define UART_LSR_SPECIAL	0x1E

#define	UART_CLK_DIV		921600

#define MX_LOCK_INIT()		unsigned long flags=0
#define MX_LOCK(lock)		{\
				if(!in_interrupt())\
					spin_lock_irqsave(lock, flags);\
				}
#define MX_UNLOCK(lock)		{\
				if(!in_interrupt())\
					spin_unlock_irqrestore(lock, flags);\
				}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
#define PORTNO(x)	(MINOR((x)->device) - (x)->driver.minor_start)
#else
#define PORTNO(x)	((x)->index)
#endif

#define RELEVANT_IFLAG(iflag)	(iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK|IXON|IXOFF))

#define IRQ_T(info) ((info->flags & ASYNC_SHARE_IRQ) ? SA_SHIRQ : SA_INTERRUPT)

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 *	Define the Moxa device IDs.
 */

#define CA104_ID    	1
#define CA132_ID    	2
#define CA132I_ID	3
#define	CA108_ID			4
#define	CA114_ID			5
#define	CA134I_ID			6

enum	{
	MXPCDRV_BOARD_CA104 = 1,
	MXPCDRV_BOARD_CA132,
	MXPCDRV_BOARD_CA132I,
	MXPCDRV_BOARD_CA108,
	MXPCDRV_BOARD_CA114,
	MXPCDRV_BOARD_CA134I
};

static char *mxpcdrv_brdname[] = {
	"CA-104 series",
	"CA-132 series",
	"CA-132I series",
	"CA-108 series",
	"CA-114 series",
	"CA-134I series"
};

static int mxpcdrv_numports[] = {
	4,	// CA104
	2,	// CA132
	2,	// CA132I
	8,	// CA108
	4,	// CA114
	4		// CA134I
};
	
/*
 *	MOXA ioctls
 */
#define MOXA			0x400
#define MOXA_GETDATACOUNT     (MOXA + 23)
#define	MOXA_GET_CONF         (MOXA + 35)
#define MOXA_DIAGNOSE         (MOXA + 50)
#define MOXA_CHKPORTENABLE    (MOXA + 60)
#define MOXA_HighSpeedOn      (MOXA + 61)
#define MOXA_GET_MAJOR        (MOXA + 63)
#define MOXA_GET_CUMAJOR      (MOXA + 64)
#define MOXA_GETMSTATUS       (MOXA + 65)
#define MOXA_SET_SPECIAL_BAUD_RATE	(MOXA + 66)
#define MOXA_GET_SPECIAL_BAUD_RATE	(MOXA + 67)

typedef struct _moxa_pci_info {
	unsigned short busNum;
	unsigned short devNum;
} moxa_pci_info;

static int ioaddr[MXPCDRV_BOARDS]={0,0,0,0};
static int iovect[MXPCDRV_BOARDS]={0,0,0,0};
static int irq[MXPCDRV_BOARDS]={0,0,0,0};
static int ttymajor=MXPCDRVMAJOR;
static int calloutmajor=MXPCDRVCUMAJOR;
static int verbose=0;

#ifdef MODULE
/* Variables for insmod */

# if (LINUX_VERSION_CODE > VERSION_CODE(2,1,11))
MODULE_AUTHOR("Eric Lo");
MODULE_DESCRIPTION("MOXA PC104 Communication Module Device Driver");
#if defined SP2 || (LINUX_VERSION_CODE > VERSION_CODE(2,6,9))
int mx_ioaddr_array_num;
int mx_iovect_array_num;
int mx_irq_array_num;
module_param_array(ioaddr, int, &mx_ioaddr_array_num, 0);
module_param_array(iovect, int, &mx_iovect_array_num, 0);
module_param_array(irq, int, &mx_irq_array_num, 0);
module_param(ttymajor, int, 0);
module_param(calloutmajor, int, 0);
module_param(verbose, int, 0);
#else
MODULE_PARM(ioaddr,     "1-4i");
MODULE_PARM(iovect,     "1-4i");
MODULE_PARM(irq,     	"1-4i");
MODULE_PARM(ttymajor,   "i");
MODULE_PARM(calloutmajor,    "i");
MODULE_PARM(verbose,        "i");
#endif
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
# endif

#endif /* MODULE */

struct mxpcdrv_hwconf {
	int		board_type;
	int		ports;
	int		irq;
	int		vector;
	int		vector_mask;
	int		uart_type;
	int		ioaddr[MXPCDRV_PORTS_PER_BOARD];
	int		baud_base[MXPCDRV_PORTS_PER_BOARD];
	moxa_pci_info	pciInfo;
	int		IsMoxaMustChipFlag;	// add by Victor Yu. 08-30-2002
	int		MaxCanSetBaudRate[MXPCDRV_PORTS_PER_BOARD];	// add by Victor Yu. 09-04-2002
};

struct mxpcdrv_struct {
	int			port;
	int			base;		/* port base address */
	int			irq;		/* port using irq no. */
	int			vector; 	/* port irq vector */
	int			vectormask;	/* port vector mask */
	int			rx_trigger;	/* Rx fifo trigger level */
	int			baud_base;	/* max. speed */
	int			flags;		/* defined in tty.h */
	int			type;		/* UART type */
	struct tty_struct *	tty;
	int			read_status_mask;
	int			ignore_status_mask;
	int			xmit_fifo_size;
	int			custom_divisor;
	int			x_char; 	/* xon/xoff character */
	int			close_delay;
	unsigned short		closing_wait;
	int			IER;		/* Interrupt Enable Register */
	int			MCR;		/* Modem control register */
	unsigned long		event;
	int			count;		/* # of fd on device */
	int			blocked_open;	/* # of blocked opens */
	long			session;	/* Session of opening process */
	long			pgrp;		/* pgrp of opening process */
	unsigned char		*xmit_buf;
	int			xmit_head;
	int			xmit_tail;
	int			xmit_cnt;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))	
	struct tq_struct	tqueue;
#else
	struct work_struct tqueue;
#endif
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
	struct ktermios		normal_termios;
	struct ktermios		callout_termios;
#else
	struct termios		normal_termios;
	struct termios		callout_termios;
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(2,4,0))
	struct wait_queue	*open_wait;
	struct wait_queue	*close_wait;
	struct wait_queue	*delta_msr_wait;
#else
	wait_queue_head_t open_wait;
	wait_queue_head_t close_wait;
	wait_queue_head_t delta_msr_wait;
#endif
	struct async_icount	icount; 	/* kernel counters for the 4 input interrupts */
	int			timeout;
	int			IsMoxaMustChipFlag;	// add by Victor Yu. 08-30-2002
	int			MaxCanSetBaudRate;	// add by Victor Yu. 09-04-2002
	unsigned char       	ldisc_stop_rx;
	spinlock_t		slock;
	int 			cur_baud;
};

struct mxpcdrv_log {
	int	tick;
	int	rxcnt[MXPCDRV_PORTS];
	int	txcnt[MXPCDRV_PORTS];
};

struct mxpcdrv_mstatus{
       tcflag_t	cflag;
       int  	cts;
       int  	dsr;
       int  	ri;
       int  	dcd;
};

static  struct mxpcdrv_mstatus GMStatus[MXPCDRV_PORTS];

static int mxpcdrvBoardIO[MXPCDRV_BOARDS]  = {
	0,0,0,0	
       /*  0x180, 0x280, 0x200, 0x320   */
};

static int mxpcdrvBoardIRQ[MXPCDRV_BOARDS]  = {
	0,0,0,0	
};

static int mxpcdrvBoardVECT[MXPCDRV_BOARDS]  = {
	0,0,0,0	
};

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
static struct tty_driver	mxvar_sdriver;
static struct tty_driver	mxvar_cdriver;
#else
static struct tty_driver	*mxvar_sdriver;
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
static int			mxvar_refcount;
#endif

//static struct tty_driver	mxvar_sdriver, mxvar_cdriver;
//static int			mxvar_refcount;
static struct mxpcdrv_struct	mxvar_table[MXPCDRV_PORTS];
static struct tty_struct *	mxvar_tty[MXPCDRV_PORTS+1];
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static struct ktermios * 	mxvar_termios[MXPCDRV_PORTS+1];
static struct ktermios * 	mxvar_termios_locked[MXPCDRV_PORTS+1];
#else
static struct termios * 	mxvar_termios[MXPCDRV_PORTS+1];
static struct termios * 	mxvar_termios_locked[MXPCDRV_PORTS+1];
#endif
static struct mxpcdrv_log 	mxvar_log;
static int			mxvar_diagflag;
/*
 * mxvar_tmp_buf is used as a temporary buffer by serial_write. We need
 * to lock it in case the memcpy_fromfs blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
static unsigned char *		mxvar_tmp_buf;
static struct semaphore 	mxvar_tmp_buf_sem;
#else
static unsigned char *		mxvar_tmp_buf = 0;
static struct semaphore 	mxvar_tmp_buf_sem = MUTEX;
#endif

/*
 * This is used to figure out the divisor speeds and the timeouts
 */
static int mxvar_baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600, 0 };
	
struct mxpcdrv_hwconf mxpcdrvcfg[MXPCDRV_BOARDS];

/*
 * static functions:
 */

#ifdef MODULE
int		init_module(void);
void		cleanup_module(void);
#endif

static void 	mxpcdrv_getcfg(int board,struct mxpcdrv_hwconf *hwconf);
int		mxpcdrv_init(void);
static int	mxpcdrv_get_ISA_conf(int, int, struct mxpcdrv_hwconf *, int);
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static void	mxpcdrv_do_softint(struct work_struct *work);
#else
static void	mxpcdrv_do_softint(void *);
#endif
static int	mxpcdrv_open(struct tty_struct *, struct file *);
static void	mxpcdrv_close(struct tty_struct *, struct file *);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
static int	mxpcdrv_write(struct tty_struct *, int, const unsigned char *, int);
#else
static int	mxpcdrv_write(struct tty_struct *, const unsigned char *, int);
#endif
static int	mxpcdrv_write_room(struct tty_struct *);
static void	mxpcdrv_flush_buffer(struct tty_struct *);
static int	mxpcdrv_chars_in_buffer(struct tty_struct *);
static void	mxpcdrv_flush_chars(struct tty_struct *);
static void	mxpcdrv_put_char(struct tty_struct *, unsigned char);
static int	mxpcdrv_ioctl(struct tty_struct *, struct file *, uint, ulong);
static int	mxpcdrv_ioctl_special(unsigned int, unsigned long);
static void	mxpcdrv_throttle(struct tty_struct *);
static void	mxpcdrv_unthrottle(struct tty_struct *);
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static void	mxpcdrv_set_termios(struct tty_struct *, struct ktermios *);
#else
static void	mxpcdrv_set_termios(struct tty_struct *, struct termios *);
#endif
static void	mxpcdrv_stop(struct tty_struct *);
static void	mxpcdrv_start(struct tty_struct *);
static void	mxpcdrv_hangup(struct tty_struct *);
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static irqreturn_t mxpcdrv_interrupt(int irq, void *dev_id);
#else
static IRQ_RET	mxpcdrv_interrupt(int, void *, struct pt_regs *);
#endif
static void mxpcdrv_receive_chars(struct mxpcdrv_struct *, int *);
static void mxpcdrv_transmit_chars(struct mxpcdrv_struct *);
static void mxpcdrv_check_modem_status(struct mxpcdrv_struct *, int);
static int	mxpcdrv_block_til_ready(struct tty_struct *, struct file *, struct mxpcdrv_struct *);
static int	mxpcdrv_startup(struct mxpcdrv_struct *);
static void	mxpcdrv_shutdown(struct mxpcdrv_struct *);
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static int	mxpcdrv_change_speed(struct mxpcdrv_struct *, struct ktermios *old_termios);
#else
static int	mxpcdrv_change_speed(struct mxpcdrv_struct *, struct termios *old_termios);
#endif
static int	mxpcdrv_get_serial_info(struct mxpcdrv_struct *, struct serial_struct *);
static int	mxpcdrv_set_serial_info(struct mxpcdrv_struct *, struct serial_struct *);
static int	mxpcdrv_get_lsr_info(struct mxpcdrv_struct *, unsigned int *);
static void	mxpcdrv_send_break(struct mxpcdrv_struct *, int);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static int mxpcdrv_tiocmget(struct tty_struct *, struct file *);
static int mxpcdrv_tiocmset(struct tty_struct *, struct file *, unsigned int, unsigned int);
#else
static int	mxpcdrv_get_modem_info(struct mxpcdrv_struct *, unsigned int *);
static int	mxpcdrv_set_modem_info(struct mxpcdrv_struct *, unsigned int, unsigned int *);
#endif
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
static void mxpcdrv_wait_until_sent(struct tty_struct *tty, int timeout);
#endif

static void mxpcdrv_startrx(struct tty_struct * tty);
static void mxpcdrv_stoprx(struct tty_struct * tty);
static void mxpcdrv_set_special_baudrate(struct mxpcdrv_struct * info, int baud);

#define READ_MOXA_MUST_GDL(baseio)	inb((baseio)+MOXA_MUST_GDL_REGISTER)

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static struct tty_operations mxpcdrv_ops = {
	.open = mxpcdrv_open,
	.close = mxpcdrv_close,
	.write = mxpcdrv_write,
	.put_char = mxpcdrv_put_char,
	.flush_chars = mxpcdrv_flush_chars,
	.write_room = mxpcdrv_write_room,
	.chars_in_buffer = mxpcdrv_chars_in_buffer,
	.flush_buffer = mxpcdrv_flush_buffer,
	.ioctl = mxpcdrv_ioctl,
	.throttle = mxpcdrv_throttle,
	.unthrottle = mxpcdrv_unthrottle,
	.set_termios = mxpcdrv_set_termios,
	.stop = mxpcdrv_stop,
	.start = mxpcdrv_start,
	.hangup = mxpcdrv_hangup,
	.tiocmget = mxpcdrv_tiocmget,
	.tiocmset = mxpcdrv_tiocmset,
};
#endif

static int CheckIsMoxaMust(int io)
{
	UCHAR	oldmcr, hwid;

	outb(0, io+UART_LCR);
	DISABLE_MOXA_MUST_ENCHANCE_MODE(io);
	oldmcr = inb(io+UART_MCR);
	outb(0, io+UART_MCR);
	SET_MOXA_MUST_XON1_VALUE(io, 0x11);
	if ( (hwid=inb(io+UART_MCR)) != 0 ) {
		outb(oldmcr, io+UART_MCR);
		return(0);
	}
	GET_MOXA_MUST_HARDWARE_ID(io, &hwid);
	if ( hwid != MOXA_MUST_HARDWARE_ID )
		return(0);
	return(1);
}
// above is modified by Victor Yu. 08-15-2002

/*
 * The MOXA PC104 Communication Module serial driver boot-time initialization code!
 */
INIT_FUNC_RET	INIT_FUNC(void)
{
	int	ret;

	if (verbose)
		printk("Loading module mxpcdrv ...\n");
	ret = mxpcdrv_init();
	if (verbose)
		printk("Done.\n");
	return (ret);
}

CLEAR_FUNC_RET	CLEAR_FUNC(void)
{
	int i,err = 0;

	if (verbose)
		printk("Unloading module mxpcdrv ...\n");

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	if ((err |= tty_unregister_driver(&mxvar_cdriver)))
		printk("Couldn't unregister MOXA PC104 Communication Module callout driver\n");
#endif		
	if ((err |= tty_unregister_driver(DRV_VAR)))
		printk("Couldn't unregister MOXA PC104 Communication Module serial driver\n");

        for(i=0; i<MXPCDRV_BOARDS; i++){
	    if(mxpcdrvcfg[i].board_type == -1)
	        continue;
            else{
	        free_irq(mxpcdrvcfg[i].irq, &mxvar_table[i*MXPCDRV_PORTS_PER_BOARD]);
		release_region(mxpcdrvcfg[i].ioaddr[0],8*mxpcdrvcfg[i].ports);
		if ((mxpcdrvcfg[i].pciInfo.busNum == 0)&&(mxpcdrvcfg[i].pciInfo.devNum == 0))
			release_region(mxpcdrvcfg[i].vector,1);
		else
			release_region(mxpcdrvcfg[i].vector,16);
            }
        }
        
	if (verbose)
		printk("Done.\n");

}

int mxpcdrv_initbrd(int board,struct mxpcdrv_hwconf *hwconf)
{
	struct mxpcdrv_struct *	info;
    int     retval;
	int	i,n;

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
	init_MUTEX(&mxvar_tmp_buf_sem); 
#endif

	n = board*MXPCDRV_PORTS_PER_BOARD;
	info = &mxvar_table[n];
	
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
		printk("        ttyN%d/cum%d - ttyN%d/cum%d  ", 
			n, n, n+hwconf->ports-1, n+hwconf->ports-1);
#else
		printk("        ttyN%d - ttyN%d ", 
			n, n+hwconf->ports-1);
#endif			
		printk(" max. baud rate = %d bps.\n", hwconf->MaxCanSetBaudRate[0]);
	
	
	for ( i=0; i<hwconf->ports; i++, n++, info++ ) {
		info->port = n;
		info->base = hwconf->ioaddr[i];
		info->irq = hwconf->irq;
		info->vector = hwconf->vector;
		info->vectormask = hwconf->vector_mask;
        info->ldisc_stop_rx = 0;
        
		// following add by Victor Yu. 08-30-2002
		// Moxa Must UART support FIFO is 64bytes for Tx/Rx
		// but receive FIFO just can set up to 62 will be OK.
		info->IsMoxaMustChipFlag = hwconf->IsMoxaMustChipFlag;
		if ( hwconf->IsMoxaMustChipFlag )
			info->rx_trigger = 56;
		else
		// above add by Victor Yu. 08-30-2002
			info->rx_trigger = 14;
		info->baud_base = hwconf->baud_base[i];
		info->flags = ASYNC_SHARE_IRQ;
		info->type = hwconf->uart_type;
		if ( (info->type == PORT_16450) || (info->type == PORT_8250) )
		    info->xmit_fifo_size = 1;
		else
		    info->xmit_fifo_size = 16;

		// following add by Victor Yu. 08-30-2002
		if ( info->IsMoxaMustChipFlag ) {
		    info->xmit_fifo_size = 64;
		    ENABLE_MOXA_MUST_ENHANCE_MODE(info->base);
		}
		info->MaxCanSetBaudRate = hwconf->MaxCanSetBaudRate[i];
		// above add by Victor Yu. 08-30-2002

		info->custom_divisor = hwconf->baud_base[i] * 16;
		info->close_delay = 5*HZ/10;
		info->closing_wait = 30*HZ;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
		INIT_WORK(&info->tqueue, mxpcdrv_do_softint);
#else
		INIT_WORK(&info->tqueue, mxpcdrv_do_softint, info);
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
		info->callout_termios = mxvar_cdriver.init_termios;
#endif
		info->normal_termios = DRV_VAR_P(init_termios);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
		init_waitqueue_head(&info->open_wait);
		init_waitqueue_head(&info->close_wait);
		init_waitqueue_head(&info->delta_msr_wait);
#endif
/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
		info->icount.rx = info->icount.tx = 0;
#endif
		info->icount.cts = info->icount.dsr = 
		    info->icount.dsr = info->icount.dcd = 0;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
		info->icount.frame = info->icount.overrun = 
		    info->icount.brk = info->icount.parity = 0;
#endif
/* */
		spin_lock_init(&info->slock);
	}
	request_region(hwconf->ioaddr[0],8*hwconf->ports,"mxpcdrv(io)");
	if ((hwconf->pciInfo.busNum == 0)&&(hwconf->pciInfo.devNum == 0))
		request_region(hwconf->vector,1,"mxpcdrv(vector)");
	else
		request_region(hwconf->vector,16,"mxpcdrv(vector)");

	/*
	 * Allocate the IRQ if necessary
	 */
	
	/* before set INT ISR, disable all int */
	for(i=0; i<hwconf->ports; i++){
		outb(inb(hwconf->ioaddr[i] + UART_IER) & 0xf0, hwconf->ioaddr[i]+UART_IER);
	}
	
	n = board*MXPCDRV_PORTS_PER_BOARD;
	info = &mxvar_table[n];

        retval = request_irq(hwconf->irq, mxpcdrv_interrupt, IRQ_T(info),
				 "mxpcdrv", info);
	if ( retval ) {
	    printk("Board %d: %s", board, mxpcdrv_brdname[hwconf->board_type-1]);
	    printk("  Request irq fail,IRQ (%d) may be conflit with another device.\n",info->irq);
	    return(retval);
	}

        return 0;
}


static void mxpcdrv_getcfg(int board,struct mxpcdrv_hwconf *hwconf)
{
	mxpcdrvcfg[board] = *hwconf;
}

int mxpcdrv_init(void)
{
	int			i, m, retval, b;
    int         ret1, ret2;
	
	struct mxpcdrv_hwconf	hwconf;
	

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
	mxvar_sdriver = alloc_tty_driver(MXPCDRV_PORTS + 1);
	if (!mxvar_sdriver)
		return -ENOMEM;
#endif

	printk("MOXA PC104 Communication Module driver version %s\n",MXPCDRV_VERSION);

	/* Initialize the tty_driver structure */
	memset(DRV_VAR, 0, sizeof(struct tty_driver));
	DRV_VAR_P(magic) = TTY_DRIVER_MAGIC;
	DRV_VAR_P(name) = "ttyN";
	DRV_VAR_P(major) = ttymajor;
	DRV_VAR_P(minor_start) = 0;
	DRV_VAR_P(num) = MXPCDRV_PORTS + 1;
	DRV_VAR_P(type) = TTY_DRIVER_TYPE_SERIAL;
	DRV_VAR_P(subtype) = SERIAL_TYPE_NORMAL;
	DRV_VAR_P(init_termios) = tty_std_termios;
	DRV_VAR_P(init_termios.c_cflag) = B9600|CS8|CREAD|HUPCL|CLOCAL;
	DRV_VAR_P(flags) = TTY_DRIVER_REAL_RAW;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
	tty_set_operations(DRV_VAR, &mxpcdrv_ops);
	DRV_VAR_P(ttys) = mxvar_tty;
#else	
	DRV_VAR_P(refcount) = &mxvar_refcount;
	DRV_VAR_P(table) = mxvar_tty;
#endif	
	DRV_VAR_P(termios) = mxvar_termios;
	DRV_VAR_P(termios_locked) = mxvar_termios_locked;

	DRV_VAR_P(open) = mxpcdrv_open;
	DRV_VAR_P(close) = mxpcdrv_close;
	DRV_VAR_P(write) = mxpcdrv_write;
	DRV_VAR_P(put_char) = mxpcdrv_put_char;
	DRV_VAR_P(flush_chars) = mxpcdrv_flush_chars;
	DRV_VAR_P(write_room) = mxpcdrv_write_room;
	DRV_VAR_P(chars_in_buffer) = mxpcdrv_chars_in_buffer;
	DRV_VAR_P(flush_buffer) = mxpcdrv_flush_buffer;
	DRV_VAR_P(ioctl) = mxpcdrv_ioctl;
	DRV_VAR_P(throttle) = mxpcdrv_throttle;
	DRV_VAR_P(unthrottle) = mxpcdrv_unthrottle;
	DRV_VAR_P(set_termios) = mxpcdrv_set_termios;
	DRV_VAR_P(stop) = mxpcdrv_stop;
	DRV_VAR_P(start) = mxpcdrv_start;
	DRV_VAR_P(hangup) = mxpcdrv_hangup;

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	DRV_VAR_P(wait_until_sent) = mxpcdrv_wait_until_sent;
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	mxvar_cdriver = mxvar_sdriver;
	mxvar_cdriver.name = "cun";
	mxvar_cdriver.major = calloutmajor;
	mxvar_cdriver.subtype = SERIAL_TYPE_CALLOUT;

	printk("Tty devices major number = %d, callout devices major number = %d\n",ttymajor,calloutmajor);
#endif

	mxvar_diagflag = 0;
	memset(mxvar_table, 0, MXPCDRV_PORTS * sizeof(struct mxpcdrv_struct));
	memset(&mxvar_log, 0, sizeof(struct mxpcdrv_log));

	m = 0;
	/* Start finding ISA boards here */
	for ( b=0; b<MXPCDRV_BOARDS && m<MXPCDRV_BOARDS; b++ ) {
            int cap,vector,irq;
	    
	    if ( !(cap=mxpcdrvBoardIO[b]) )
	        continue;
	    if ( !(vector=mxpcdrvBoardVECT[b]) )
	        continue;
	    if ( !(irq=mxpcdrvBoardIRQ[b]) )
	        continue;
        
	    retval = mxpcdrv_get_ISA_conf(cap, vector, &hwconf, irq);

	    if ( retval != 0 ) 
	    	printk("Found MOXA %s board (IO=0x%x,Vector=0x%x,IRQ=%d)\n",
				mxpcdrv_brdname[hwconf.board_type-1],
				mxpcdrvBoardIO[b],mxpcdrvBoardVECT[b],
				mxpcdrvBoardIRQ[b]);
				
	    if ( retval <= 0 ) {
		if (retval == MXPCDRV_ERR_IRQ)
			printk("Invalid interrupt number,board not configured\n");
		else if (retval == MXPCDRV_ERR_IRQ_CONFLIT)
			printk("Invalid interrupt number,board not configured\n");
		else if (retval == MXPCDRV_ERR_VECTOR)
			printk("Invalid interrupt vector,board not configured\n");
		else if (retval == MXPCDRV_ERR_IOADDR)
			printk("Invalid I/O address,board not configured\n");

		continue;
	    }
	    
	    hwconf.pciInfo.busNum = 0;
	    hwconf.pciInfo.devNum = 0;
	    mxpcdrv_getcfg(m,&hwconf);
	    
	    if(mxpcdrv_initbrd(m,&hwconf)<0)
                continue;
	    
				  
	    m++;
	}
	/* Start finding ISA boards from module arg */
	for ( b=0; b<MXPCDRV_BOARDS && m<MXPCDRV_BOARDS; b++ ) {
            int cap,vector;

	    if ( !(cap=ioaddr[b]) )
	        continue;
	    if ( !(vector=iovect[b]) )
		continue;
	    if ( !(irq[b]) )
		continue;
                
	    retval = mxpcdrv_get_ISA_conf(cap, vector, &hwconf, irq[b]);

	    if ( retval != 0 ) 
	    	printk("Found MOXA %s board (IO=0x%x,Vector=0x%x,IRQ=%d)\n",
				mxpcdrv_brdname[hwconf.board_type-1],
				ioaddr[b],iovect[b],irq[b]);
				
	    if ( retval <= 0 ) {
		if (retval == MXPCDRV_ERR_IRQ)
			printk("Invalid interrupt number,board not configured\n");
		else if (retval == MXPCDRV_ERR_IRQ_CONFLIT)
			printk("Invalid interrupt number,board not configured\n");
		else if (retval == MXPCDRV_ERR_VECTOR)
			printk("Invalid interrupt vector,board not configured\n");
		else if (retval == MXPCDRV_ERR_IOADDR)
			printk("Invalid I/O address,board not configured\n");

		continue;
	    }
	    
	    hwconf.pciInfo.busNum = 0;
	    hwconf.pciInfo.devNum = 0;
	    
	    mxpcdrv_getcfg(m,&hwconf);

	    if(mxpcdrv_initbrd(m,&hwconf)<0)
                continue;
	    
	    m++;
	}

	for(i=m; i<MXPCDRV_BOARDS; i++){
		mxpcdrvcfg[i].board_type = -1;
	}

        ret1 = 0;
        ret2 = 0;
        if ( !(ret1=tty_register_driver(DRV_VAR)) ){
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))        	
            if ( !(ret2=tty_register_driver(&mxvar_cdriver)) ){
                return 0;
            }else{
                tty_unregister_driver(DRV_VAR);
		printk("Couldn't install MOXA PC104 Communication Module callout driver !\n");
            }
#else
    		return 0;	            
#endif
        }else
	    printk("Couldn't install MOXA PC104 Communication Module driver !\n");


        if(ret1 || ret2){
            for(i=0; i<MXPCDRV_BOARDS; i++){
	        if(mxpcdrvcfg[i].board_type == -1)
	            continue;
                else{
		    free_irq(mxpcdrvcfg[i].irq, &mxvar_table[i*MXPCDRV_PORTS_PER_BOARD]);
                }
            }
            return -1;
        }
        
	return(0);
}

#if (LINUX_VERSION_CODE >  VERSION_CODE(2,6,19))
static void mxpcdrv_do_softint(struct work_struct *work)
{
	struct mxpcdrv_struct *	info = container_of(work,struct mxpcdrv_struct, tqueue);
#else
static void mxpcdrv_do_softint(void *private_)
{
	struct mxpcdrv_struct *	info = (struct mxpcdrv_struct *)private_;
#endif
	struct tty_struct *	tty;
	
	tty = info->tty;

#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
	    if ( clear_bit(MXPCDRV_EVENT_TXLOW, &info->event) ) {
	        if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		 tty->ldisc.write_wakeup )
		(tty->ldisc.write_wakeup)(tty);
	       wake_up_interruptible(&tty->write_wait);
	    }
	    if ( clear_bit(MXPCDRV_EVENT_HANGUP, &info->event) ) {
		tty_hangup(tty);
	    }
#else
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,9))
	    if ( test_and_clear_bit(MXPCDRV_EVENT_TXLOW, &info->event) ) {
	        if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		 tty->ldisc.write_wakeup )
		(tty->ldisc.write_wakeup)(tty);
	        wake_up_interruptible(&tty->write_wait);
	    }
#else
	    if ( test_and_clear_bit(MXPCDRV_EVENT_TXLOW, &info->event) ) 
	    	tty_wakeup(tty);
#endif	    	    
	    if ( test_and_clear_bit(MXPCDRV_EVENT_HANGUP, &info->event) ) {
		tty_hangup(tty);
	    }
#endif 
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	MX_MOD_DEC;  
#endif
}

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its async structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
static int mxpcdrv_open(struct tty_struct * tty, struct file * filp)
{
	struct mxpcdrv_struct *	info;
	int			retval, line;
	unsigned long		page;

	line = PORTNO(tty);
	if ( line == MXPCDRV_PORTS )
	    return(0);
	
	MX_MOD_INC;

	if ( (line < 0) || (line > MXPCDRV_PORTS) )
	    return(-ENODEV);
	info = mxvar_table + line;
	if ( !info->base )
	    return(-ENODEV);

	tty->driver_data = info;
	info->tty = tty;

	if ( !mxvar_tmp_buf ) {
	    page = GET_FPAGE(GFP_KERNEL);
	    if ( !page )
		return(-ENOMEM);
	    if ( mxvar_tmp_buf )
		free_page(page);
	    else
		mxvar_tmp_buf = (unsigned char *)page;
	}

	/*
	 * Start up serial port
	 */
	info->count++;

	retval = mxpcdrv_startup(info);
	if ( retval )
	    return(retval);

	retval = mxpcdrv_block_til_ready(tty, filp, info);
	if ( retval )
	    return(retval);


	if ( (info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS) ) {
	    if ( MX_TTY_DRV(subtype) == SERIAL_TYPE_NORMAL )
		*tty->termios = info->normal_termios;
	    else
		*tty->termios = info->callout_termios;
	    mxpcdrv_change_speed(info, 0);
	}

	info->session = MX_SESSION();
	info->pgrp = MX_CGRP();
#ifdef TTY_DONT_FLIP
	clear_bit(TTY_DONT_FLIP, &tty->flags);
#endif

/* unmark here for very high baud rate (ex. 921600 bps) used
*/
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	tty->low_latency = 1;
#endif
	return(0);
}

/*
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 */
static void mxpcdrv_close(struct tty_struct * tty, struct file * filp)
{
	struct mxpcdrv_struct * info = (struct mxpcdrv_struct *)tty->driver_data;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,9))
	struct tty_ldisc *ld;
#endif
	unsigned long	timeout;

	MX_LOCK_INIT();
	if ( PORTNO(tty) == MXPCDRV_PORTS )
	    return;
	if ( !info ){
			MX_MOD_DEC;
	    return;
	}

	MX_LOCK(&info->slock);
	if ( tty_hung_up_p(filp) ) {
	    MX_UNLOCK(&info->slock);
	    MX_MOD_DEC;
	    return;
	}

#ifndef SP1
	if ( (tty->count == 1) && (info->count != 1) ) {
#else
#if (LINUX_VERSION_CODE < VERSION_CODE(2,4,21))
	if ( (tty->count == 1) && (info->count != 1) ) {
#else
	if ((atomic_read(&tty->count) == 1) && (info->count != 1)) {
#endif
#endif
	    /*
	     * Uh, oh.	tty->count is 1, which means that the tty
	     * structure will be freed.  Info->count should always
	     * be one in these conditions.  If it's greater than
	     * one, we've got real problems, since it means the
	     * serial port won't be shutdown.
	     */
	    printk("mxpcdrv_close: bad serial port count; tty->count is 1, "
		   "info->count is %d\n", info->count);
	    info->count = 1;
	}
	if ( --info->count < 0 ) {
	    printk("mxpcdrv_close: bad serial port count for ttys%d: %d\n",
		   info->port, info->count);
	    info->count = 0;
	}
	if ( info->count ) {
	    MX_UNLOCK(&info->slock);
	    MX_MOD_DEC;
	    return;
	}
	info->flags |= ASYNC_CLOSING;
	MX_UNLOCK(&info->slock);	// add by Victor Yu. 09-26-2002
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if ( info->flags & ASYNC_NORMAL_ACTIVE )
	    info->normal_termios = *tty->termios;
	if ( info->flags & ASYNC_CALLOUT_ACTIVE )
	    info->callout_termios = *tty->termios;
	/*
	 * Now we wait for the transmit buffer to clear; and we notify
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if ( info->closing_wait != ASYNC_CLOSING_WAIT_NONE )
	    tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	info->IER &= ~UART_IER_RLSI;
	if ( info->IsMoxaMustChipFlag )
		info->IER &= ~MOXA_MUST_RECV_ISR;
/* by William
	info->read_status_mask &= ~UART_LSR_DR;
*/
	if ( info->flags & ASYNC_INITIALIZED ) {
	    outb(info->IER, info->base + UART_IER);
	    /*
	     * Before we drop DTR, make sure the UART transmitter
	     * has completely drained; this is especially
	     * important if there is a transmit FIFO!
	     */
	    timeout = jiffies + HZ;
	    while ( !(inb(info->base + UART_LSR) & UART_LSR_TEMT) ) {
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
				set_current_state(TASK_INTERRUPTIBLE);
#else
				current->state = TASK_INTERRUPTIBLE;
#endif
		schedule_timeout(5);
		if ( time_after(jiffies, timeout) )
		    break;
	    }
	}
	mxpcdrv_shutdown(info);

	// following add by Victor Yu. 09-23-2002
	/*
	if ( info->IsMoxaMustChipFlag )
		SET_MOXA_MUST_NO_SOFTWARE_FLOW_CONTROL(info->base);
	*/
	// above add by Victor Yu. 09-23-2002

	if ( MX_TTY_DRV(flush_buffer) )
	    MX_TTY_DRV(flush_buffer)(tty);
	    
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
	if ( tty->ldisc.flush_buffer )
	    tty->ldisc.flush_buffer(tty);
#else    
	ld = tty_ldisc_ref(tty);
	if (ld) {
		if(ld->flush_buffer)
			ld->flush_buffer(tty);
		tty_ldisc_deref(ld);
	}
#endif

	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if ( info->blocked_open ) {
	    if ( info->close_delay ) {
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
				set_current_state(TASK_INTERRUPTIBLE);
#else
				current->state = TASK_INTERRUPTIBLE;
#endif
		schedule_timeout(info->close_delay);
	    }
	    wake_up_interruptible(&info->open_wait);
	}

	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE |
			 ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
//	MX_UNLOCK(&info->slock);	// mask by Vicror Yu. 09-26-2002

	MX_MOD_DEC;
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
static int mxpcdrv_write(struct tty_struct * tty, int from_user,
		       const unsigned char * buf, int count)
#else
static int mxpcdrv_write(struct tty_struct * tty, 
		       const unsigned char * buf, int count)
#endif
{
	int		c, total = 0;
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

	if ( !tty || !info->xmit_buf || !mxvar_tmp_buf )
	    return(0);

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
	if ( from_user )
	    down(&mxvar_tmp_buf_sem);
#endif

	while ( 1 ) {
	    c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
			       SERIAL_XMIT_SIZE - info->xmit_head));
	    if ( c <= 0 )
		break;

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
	    if ( from_user ) {
		/*
		copy_from_user(mxvar_tmp_buf, buf, c);
		c = MIN(c, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
			       SERIAL_XMIT_SIZE - info->xmit_head));
		memcpy(info->xmit_buf + info->xmit_head, mxvar_tmp_buf, c);
	    */
		if(copy_from_user(info->xmit_buf+info->xmit_head, buf, c)==c){
		    total = -EFAULT;	
		    break;
		}
	    } else
		memcpy(info->xmit_buf + info->xmit_head, buf, c);
#else
		memcpy(info->xmit_buf + info->xmit_head, buf, c);
#endif
	    MX_LOCK(&info->slock);    
	    info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE - 1);
	    info->xmit_cnt += c;
	    MX_UNLOCK(&info->slock);
	    buf += c;
	    count -= c;
	    total += c;
	}
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
	if ( from_user )
	    up(&mxvar_tmp_buf_sem);
#endif

	if ( info->xmit_cnt && !tty->stopped ){ 
        	    MX_LOCK(&info->slock);
			        info->IER &= ~UART_IER_THRI;
	            outb(info->IER, info->base + UART_IER);
	            info->IER |= UART_IER_THRI;
	            outb(info->IER, info->base + UART_IER);
	            MX_UNLOCK(&info->slock);
	}

	return(total);
}

static void mxpcdrv_put_char(struct tty_struct * tty, unsigned char ch)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

	if ( !tty || !info->xmit_buf )
	    return;

	if ( info->xmit_cnt >= SERIAL_XMIT_SIZE - 1 ) {
	    return;
	}

	MX_LOCK(&info->slock);
	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= SERIAL_XMIT_SIZE - 1;
	info->xmit_cnt++;
	MX_UNLOCK(&info->slock);
	if ( !tty->stopped/* && !(info->IER & UART_IER_THRI) */) {
        		MX_LOCK(&info->slock);
		        info->IER &= ~UART_IER_THRI	;
            outb(info->IER, info->base + UART_IER);
		    		info->IER |= UART_IER_THRI;
		    		outb(info->IER, info->base + UART_IER);
	  	  		MX_UNLOCK(&info->slock);
	}
}

static void mxpcdrv_flush_chars(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

	if ( info->xmit_cnt <= 0 || tty->stopped || !info->xmit_buf || 
	     (tty->hw_stopped && (info->type!=PORT_16550A) && (!info->IsMoxaMustChipFlag)))
	    return;

	MX_LOCK(&info->slock);
  info->IER &= ~UART_IER_THRI	;
  outb(info->IER, info->base + UART_IER);
	info->IER |= UART_IER_THRI;
	outb(info->IER, info->base + UART_IER);
	MX_UNLOCK(&info->slock);
}

static int mxpcdrv_write_room(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	int	ret;

	ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
	if ( ret < 0 )
	    ret = 0;
	return(ret);
}

static int mxpcdrv_chars_in_buffer(struct tty_struct * tty)
{
	int len;
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	len = info->xmit_cnt;

	if(!(inb(info->base + UART_LSR) & UART_LSR_THRE))  
		len++;
		
	return len;
}

static void mxpcdrv_flush_buffer(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	char fcr;
	MX_LOCK_INIT();

	MX_LOCK(&info->slock);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	MX_UNLOCK(&info->slock);

	fcr = inb(info->base + UART_FCR);
	outb((fcr | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), 
		info->base + UART_FCR);
	outb(fcr, info->base+UART_FCR);

	wake_up_interruptible(&tty->write_wait);
	if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	     tty->ldisc.write_wakeup )
	    (tty->ldisc.write_wakeup)(tty);
}

static int mxpcdrv_ioctl(struct tty_struct * tty, struct file * file,
		       unsigned int cmd, unsigned long arg)
{
	int			error;
	struct mxpcdrv_struct *	info = (struct mxpcdrv_struct *)tty->driver_data;
	int			retval;
	struct async_icount	cprev, cnow;	    /* kernel counter temps */
	struct serial_icounter_struct *p_cuser;     /* user space */
	unsigned long 		templ;

	MX_LOCK_INIT();
	
	if ( PORTNO(tty) == MXPCDRV_PORTS )
	    return(mxpcdrv_ioctl_special(cmd, arg));
	if ( (cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) &&
	     (cmd != TIOCGICOUNT) ) {
	    if ( tty->flags & (1 << TTY_IO_ERROR) )
		return(-EIO);
	}
	switch ( cmd ) {
	case TCSBRK:	/* SVID version: non-zero arg --> no break */
	    retval = tty_check_change(tty);
	    if ( retval )
		return(retval);
	    tty_wait_until_sent(tty, 0);
	    if ( !arg )
		mxpcdrv_send_break(info, HZ/4);		/* 1/4 second */
	    return(0);
	case TCSBRKP:	/* support for POSIX tcsendbreak() */
	    retval = tty_check_change(tty);
	    if ( retval )
		return(retval);
	    tty_wait_until_sent(tty, 0);
	    mxpcdrv_send_break(info, arg ? arg*(HZ/10) : HZ/4);
	    return(0);
	case TIOCGSOFTCAR:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(long));
	    if ( MX_ERR(error) )
		return(error);
	    put_to_user(C_CLOCAL(tty) ? 1 : 0, (unsigned long *)arg);
	    return 0;
	case TIOCSSOFTCAR:
	    error = MX_ACCESS_CHK(VERIFY_READ, (void *)arg, sizeof(long));
	    if ( MX_ERR(error) )
		return(error);
	    get_from_user(templ,(unsigned long *)arg);
	    arg = templ;
	    tty->termios->c_cflag = ((tty->termios->c_cflag & ~CLOCAL) |
				    (arg ? CLOCAL : 0));
	    return(0);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	case TIOCMGET:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg,
				sizeof(unsigned int));
	    if ( MX_ERR(error) )
		return(error);
	    return(mxpcdrv_get_modem_info(info, (unsigned int *)arg));
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
	    return(mxpcdrv_set_modem_info(info, cmd, (unsigned int *)arg));
#endif
	case TIOCGSERIAL:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg,
				sizeof(struct serial_struct));
	    if ( MX_ERR(error) )
		return(error);
	    return(mxpcdrv_get_serial_info(info, (struct serial_struct *)arg));
	case TIOCSSERIAL:
	    error = MX_ACCESS_CHK(VERIFY_READ, (void *)arg,
				sizeof(struct serial_struct));
	    if ( MX_ERR(error) )
		return(error);
	    return(mxpcdrv_set_serial_info(info, (struct serial_struct *)arg));
	case TIOCSERGETLSR: /* Get line status register */
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg,
				sizeof(unsigned int));
	    if ( MX_ERR(error) )
		return(error);
	    else
		return(mxpcdrv_get_lsr_info(info, (unsigned int *)arg));
	/*
	 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
	 * - mask passed in arg for lines of interest
	 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
	 * Caller should use TIOCGICOUNT to see which one it was
	 */
	case TIOCMIWAIT:
	    MX_LOCK(&info->slock);
	    cprev = info->icount;   /* note the counters on entry */
	    MX_UNLOCK(&info->slock);
	    while ( 1 ) {
		//interruptible_sleep_on(&info->delta_msr_wait);
		/* see if a signal did it */
		if ( signal_pending(current) )
		    return(-ERESTARTSYS);

	    MX_LOCK(&info->slock);
		cnow = info->icount;	/* atomic copy */
        MX_UNLOCK(&info->slock);
		if ( cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
		     cnow.dcd == cprev.dcd && cnow.cts == cprev.cts )
		    return(-EIO);	/* no change => error */
		if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
		     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
		     ((arg & TIOCM_CD)	&& (cnow.dcd != cprev.dcd)) ||
		     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
		    return(0);
		}
		cprev = cnow;
	    }
	    /* NOTREACHED */
	/*
	 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
	 * Return: write counters to the user passed counter struct
	 * NB: both 1->0 and 0->1 transitions are counted except for
	 *     RI where only 0->1 is counted.
	 */
	case TIOCGICOUNT:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg,
				sizeof(struct serial_icounter_struct));
	    if ( MX_ERR(error) )
		return(error);
	    MX_LOCK(&info->slock);
	    cnow = info->icount;
	    MX_UNLOCK(&info->slock);
	    p_cuser = (struct serial_icounter_struct *)arg;
/* modified by casper 1/11/2000 */ 
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	        if (put_user(cnow.frame, &p_cuser->frame))
			return -EFAULT;
	        if (put_user(cnow.brk, &p_cuser->brk))
			return -EFAULT;
	        if (put_user(cnow.overrun, &p_cuser->overrun))
			return -EFAULT;
	        if (put_user(cnow.buf_overrun, &p_cuser->buf_overrun))
			return -EFAULT;
	        if (put_user(cnow.parity, &p_cuser->parity))
			return -EFAULT;
	        if (put_user(cnow.rx, &p_cuser->rx))
			return -EFAULT;
	        if (put_user(cnow.tx, &p_cuser->tx))
			return -EFAULT;
#endif

	    put_to_user(cnow.cts, &p_cuser->cts);
	    put_to_user(cnow.dsr, &p_cuser->dsr);
	    put_to_user(cnow.rng, &p_cuser->rng);
	    put_to_user(cnow.dcd, &p_cuser->dcd);

/* */
	    return(0);
	case MOXA_HighSpeedOn:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(int));
	    if ( MX_ERR(error) )
		return(error);
	    put_to_user(info->baud_base != 115200 ? 1 : 0, (int *)arg);
	    return(0);
	    
	case MOXA_SET_SPECIAL_BAUD_RATE:
		error = MX_ACCESS_CHK(VERIFY_READ, (void *)arg, sizeof(int));
	    	if ( MX_ERR(error) )
			return(error);
			
		MX_LOCK(&info->slock);
		mxpcdrv_set_special_baudrate(info, (int)arg);
		MX_UNLOCK(&info->slock);

		return(0);
			
	case MOXA_GET_SPECIAL_BAUD_RATE:
		error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(int));
	    	if ( MX_ERR(error) )
			return(error);
				
		put_to_user(info->cur_baud, (int *)arg);
					
		return(0);
	    
	default:
	    return(-ENOIOCTLCMD);
	}
	return(0);
}

static int mxpcdrv_ioctl_special(unsigned int cmd, unsigned long arg)
{
	int		error, i, result, status;

	switch ( cmd ) {
	case MOXA_GET_CONF:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, 
			    sizeof(struct mxpcdrv_hwconf)*4);
	    if ( MX_ERR(error) )
		return(error);
	    if (copy_to_user((struct mxpcdrv_hwconf *)arg, mxpcdrvcfg, 
			    sizeof(struct mxpcdrv_hwconf)*4))
		return -EFAULT;
	    return 0;
        case MOXA_GET_MAJOR:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(int));
	    if ( MX_ERR(error) )
                return(error);
	    if (copy_to_user((int*)arg, &ttymajor, sizeof(int)))
		return -EFAULT;
            return 0;

        case MOXA_GET_CUMAJOR:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(int));
	    if ( MX_ERR(error) )
                return(error);
	    if (copy_to_user((int*)arg, &calloutmajor, sizeof(int)))
		return -EFAULT;
            return 0;

	case MOXA_CHKPORTENABLE:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(long));
	    if ( MX_ERR(error) )
		return(error);
	    result = 0;
	    for ( i=0; i<MXPCDRV_PORTS; i++ ) {
		if ( mxvar_table[i].base )
		    result |= (1 << i);
	    }
	    put_to_user(result, (unsigned long *)arg);
	    return(0);
	case MOXA_GETDATACOUNT:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg,
				sizeof(struct mxpcdrv_log));
	    if ( MX_ERR(error) )
		return(error);
	    if (copy_to_user((struct mxpcdrv_log *)arg, &mxvar_log, sizeof(mxvar_log)))
		return -EFAULT;
	    return(0);
        case MOXA_GETMSTATUS:
	    error = MX_ACCESS_CHK(VERIFY_WRITE, (void *)arg,
				sizeof(struct mxpcdrv_mstatus) * MXPCDRV_PORTS);
	    if ( MX_ERR(error) )
		return(error);

            for(i=0; i<MXPCDRV_PORTS; i++){
                GMStatus[i].ri = 0;
		if ( !mxvar_table[i].base ){
                    GMStatus[i].dcd = 0;
                    GMStatus[i].dsr = 0;
                    GMStatus[i].cts = 0;
		    continue;
                }

	        if ( !mxvar_table[i].tty || !mxvar_table[i].tty->termios )
                    GMStatus[i].cflag=mxvar_table[i].normal_termios.c_cflag;
                else
                    GMStatus[i].cflag = mxvar_table[i].tty->termios->c_cflag;
                
                status = inb(mxvar_table[i].base + UART_MSR);
                if(status  & 0x80/*UART_MSR_DCD*/)
                    GMStatus[i].dcd = 1;
                else
                    GMStatus[i].dcd = 0;

                if(status  & 0x20/*UART_MSR_DSR*/)
                    GMStatus[i].dsr = 1;
                else
                    GMStatus[i].dsr = 0;
                

                if(status  & 0x10/*UART_MSR_CTS*/)
                    GMStatus[i].cts = 1;
                else
                    GMStatus[i].cts = 0;
            }
            if (copy_to_user((struct mxpcdrv_mstatus *)arg, GMStatus,
                                    sizeof(struct mxpcdrv_mstatus) * MXPCDRV_PORTS))
		return -EFAULT;
            return 0;
	default:
	    return(-ENOIOCTLCMD);
	}
	return(0);
}

static void mxpcdrv_stoprx(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();


        info->ldisc_stop_rx = 1;
	if ( I_IXOFF(tty) ) {

	    MX_LOCK(&info->slock);
	    // following add by Victor Yu. 09-02-2002
	    if ( info->IsMoxaMustChipFlag ) {
		info->IER &= ~MOXA_MUST_RECV_ISR;
		outb(info->IER, info->base+UART_IER);
	    } else {
	    // above add by Victor Yu. 09-02-2002
	      if(!(info->flags & ASYNC_CLOSING)){
	        info->x_char = STOP_CHAR(tty);
	//      outb(info->IER, 0); // mask by Victor Yu. 09-02-2002
	        outb(0, info->base+UART_IER);
	        info->IER |= UART_IER_THRI;
	        outb(info->IER, info->base + UART_IER);	/* force Tx interrupt */
	      }
	    }	// add by Victor Yu. 09-02-2002
	    MX_UNLOCK(&info->slock);
	}

	if ( info->tty->termios->c_cflag & CRTSCTS ) {
	    MX_LOCK(&info->slock);
	    info->MCR &= ~UART_MCR_RTS;
	    outb(info->MCR, info->base + UART_MCR);
	    MX_UNLOCK(&info->slock);
	}
}

static void mxpcdrv_startrx(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

        info->ldisc_stop_rx = 0;
        if ( I_IXOFF(tty) ) {
	    if ( info->x_char )
		info->x_char = 0;
	    else {
		MX_LOCK(&info->slock);

		// following add by Victor Yu. 09-02-2002
		if ( info->IsMoxaMustChipFlag ) {
			info->IER |= MOXA_MUST_RECV_ISR;
			outb(info->IER, info->base+UART_IER);
		} else {
		// above add by Victor Yu. 09-02-2002
			if(!(info->flags & ASYNC_CLOSING)){
		    info->x_char = START_CHAR(tty);
	//	    outb(info->IER, 0);	// mask by Victor Yu. 09-02-2002
		    outb(0, info->base+UART_IER);	// add by Victor Yu. 09-02-2002
		    info->IER |= UART_IER_THRI;		/* force Tx interrupt */
		    outb(info->IER, info->base + UART_IER);
		  }
		}	// add by Victor Yu. 09-02-2002
		MX_UNLOCK(&info->slock);
	    }
	}

	if ( info->tty->termios->c_cflag & CRTSCTS ) {
	    MX_LOCK(&info->slock);
	    info->MCR |= UART_MCR_RTS;
	    outb(info->MCR, info->base + UART_MCR);
	    MX_UNLOCK(&info->slock);
	}
}

/*
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 */
static void mxpcdrv_throttle(struct tty_struct * tty)
{
	mxpcdrv_stoprx(tty);
}

static void mxpcdrv_unthrottle(struct tty_struct * tty)
{
	mxpcdrv_startrx(tty);
}

#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static void mxpcdrv_set_termios(struct tty_struct * tty,
			      struct ktermios * old_termios)
#else
static void mxpcdrv_set_termios(struct tty_struct * tty,
			      struct termios * old_termios)
#endif
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

	if ( (tty->termios->c_cflag != old_termios->c_cflag) ||
	     (RELEVANT_IFLAG(tty->termios->c_iflag) !=
	      RELEVANT_IFLAG(old_termios->c_iflag)) ) {

		mxpcdrv_change_speed(info, old_termios);

		if ( (old_termios->c_cflag & CRTSCTS) &&
	     	     !(tty->termios->c_cflag & CRTSCTS) ) {
	    		tty->hw_stopped = 0;
	    		mxpcdrv_start(tty);
		}
	}
	
/* Handle sw stopped */
	if ( (old_termios->c_iflag & IXON) &&
     	     !(tty->termios->c_iflag & IXON) ) {
    		tty->stopped = 0;
		
		// following add by Victor Yu. 09-02-2002
		if ( info->IsMoxaMustChipFlag ) {
		    MX_LOCK(&info->slock);
			DISABLE_MOXA_MUST_RX_SOFTWARE_FLOW_CONTROL(info->base);
			MX_UNLOCK(&info->slock);
		}
		// above add by Victor Yu. 09-02-2002

    		mxpcdrv_start(tty);
	}
}

/*
 * mxpcdrv_stop() and mxpcdrv_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 */
static void mxpcdrv_stop(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

	MX_LOCK(&info->slock);
	if ( info->IER & UART_IER_THRI ) {
	    info->IER &= ~UART_IER_THRI;
	    outb(info->IER, info->base + UART_IER);
	}
	MX_UNLOCK(&info->slock);
}

static void mxpcdrv_start(struct tty_struct * tty)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *)tty->driver_data;
	MX_LOCK_INIT();

	MX_LOCK(&info->slock);	
	if ( info->xmit_cnt && info->xmit_buf/* &&
	     !(info->IER & UART_IER_THRI)*/ ) {
	    info->IER &= ~UART_IER_THRI	;
	    outb(info->IER, info->base + UART_IER);
	    info->IER |= UART_IER_THRI;
	    outb(info->IER, info->base + UART_IER);
	}
	MX_UNLOCK(&info->slock);
}

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
/*
 * mxpcdrv_wait_until_sent() --- wait until the transmitter is empty
 */
static void mxpcdrv_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct mxpcdrv_struct * info = (struct mxpcdrv_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int lsr;
	
	if (info->type == PORT_UNKNOWN)
		return;

	if (info->xmit_fifo_size == 0)
		return; /* Just in case.... */

	orig_jiffies = jiffies;
	/*
	 * Set the check interval to be 1/5 of the estimated time to
	 * send a single character, and make it at least 1.  The check
	 * interval should also be less than the timeout.
	 * 
	 * Note: we have to use pretty tight timings here to satisfy
	 * the NIST-PCTS.
	 */
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;
	if (char_time == 0)
		char_time = 1;
	if (timeout && timeout < char_time)
		char_time = timeout;
	/*
	 * If the transmitter hasn't cleared in twice the approximate
	 * amount of time to send the entire FIFO, it probably won't
	 * ever clear.  This assumes the UART isn't doing flow
	 * control, which is currently the case.  Hence, if it ever
	 * takes longer than info->timeout, this is probably due to a
	 * UART bug of some kind.  So, we clamp the timeout parameter at
	 * 2*info->timeout.
	 */
	if (!timeout || timeout > 2*info->timeout)
		timeout = 2*info->timeout;
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("In rs_wait_until_sent(%d) check=%lu...", timeout, char_time);
	printk("jiff=%lu...", jiffies);
#endif
	while (!((lsr = inb(info->base+ UART_LSR)) & UART_LSR_TEMT)) {
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
		printk("lsr = %d (jiff=%lu)...", lsr, jiffies);
#endif
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
		set_current_state(TASK_INTERRUPTIBLE);
#else
		current->state = TASK_INTERRUPTIBLE;
#endif
		schedule_timeout(char_time);
		if (signal_pending(current))
			break;
		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
		set_current_state(TASK_RUNNING);
#else
		current->state = TASK_RUNNING;
#endif

#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("lsr = %d (jiff=%lu)...done\n", lsr, jiffies);
#endif
}
#endif


/*
 * This routine is called by tty_hangup() when a hangup is signaled.
 */
void mxpcdrv_hangup(struct tty_struct * tty)
{
	struct mxpcdrv_struct * info = (struct mxpcdrv_struct *)tty->driver_data;

	mxpcdrv_flush_buffer(tty);
	mxpcdrv_shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

/*
 * This is the serial driver's generic interrupt routine
 */
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static irqreturn_t mxpcdrv_interrupt(int irq, void *dev_id)
#else
static IRQ_RET mxpcdrv_interrupt(int irq, void *dev_id, struct pt_regs * regs)
#endif
{
	int			status, iir, i;
	struct mxpcdrv_struct *	info;
	struct mxpcdrv_struct *	port;
	int			max, irqbits, bits, msr;
	int			pass_counter = 0;
	int			handled = 0;

        port = 0;
        for(i=0; i<MXPCDRV_BOARDS; i++){
            if(dev_id == &(mxvar_table[i*MXPCDRV_PORTS_PER_BOARD])){
                port = dev_id;
                break;
            }
        }

        if(i==MXPCDRV_BOARDS)
            goto irq_stop;
        if(port==0)
            goto irq_stop;
        max = mxpcdrv_numports[mxpcdrvcfg[i].board_type-1];
        
	while ( 1 ) {
	    irqbits = inb(port->vector) & port->vectormask;
	    if ( irqbits == port->vectormask )
			break;

	    handled = 1;
	    for ( i=0, bits=1; i<max; i++, irqbits |= bits, bits <<= 1 ) {
		if ( irqbits == port->vectormask )
		   	break;
		if ( bits & irqbits )
		    continue;
		info = port + i;

		// following add by Victor Yu. 09-13-2002
		iir = inb(info->base+UART_IIR);
		if ( iir & UART_IIR_NO_INT )
			continue;
		iir &= MOXA_MUST_IIR_MASK;
		if ( !info->tty ) {
			status = inb(info->base+UART_LSR);
			outb(0x27, info->base+UART_FCR);
			inb(info->base+UART_MSR);
			continue;
		}
		// above add by Victor Yu. 09-13-2002

		/* mask by Victor Yu. 09-13-2002
		if ( !info->tty ||
		     (inb(info->base + UART_IIR) & UART_IIR_NO_INT) )
		    continue;
		*/
		/* mask by Victor Yu. 09-02-2002
		status = inb(info->base + UART_LSR) & info->read_status_mask;
		*/

		// following add by Victor Yu. 09-02-2002
		status = inb(info->base+UART_LSR);
		if ( info->IsMoxaMustChipFlag ) {
			/*
			if ( (status & 0x02) && !(status & 0x01) ) {
				outb(info->base+UART_FCR,  0x23);
				continue;
			}
			*/
			if ( iir == MOXA_MUST_IIR_GDA ||
			     iir == MOXA_MUST_IIR_RDA ||
			     iir == MOXA_MUST_IIR_RTO ||
			     iir == MOXA_MUST_IIR_LSR )
				mxpcdrv_receive_chars(info, &status);
		} else {
		// above add by Victor Yu. 09-02-2002

			status &= info->read_status_mask;
			if ( status & UART_LSR_DR )
			    mxpcdrv_receive_chars(info, &status);
		}
		msr = inb(info->base + UART_MSR);
		if ( msr & UART_MSR_ANY_DELTA )
		    mxpcdrv_check_modem_status(info, msr);

		// following add by Victor Yu. 09-13-2002
		if ( info->IsMoxaMustChipFlag ) {
			if ( iir == 0x02 )
				mxpcdrv_transmit_chars(info);
		} else {
		// above add by Victor Yu. 09-13-2002

			if ( status & UART_LSR_THRE ) {
/* 8-2-99 by William
			    if ( info->x_char || (info->xmit_cnt > 0) )
*/
				mxpcdrv_transmit_chars(info);
			}
		}
	    }
	    if ( pass_counter++ > MXPCDRV_ISR_PASS_LIMIT ) {
#if 0
		printk("MOXA PC104 Communication Module driver interrupt loop break\n");
#endif
		break;	/* Prevent infinite loops */
	    }
	}
irq_stop:
	return IRQ_RETVAL(handled);
}

#if defined (_SCREEN_INFO_H) || (LINUX_VERSION_CODE > VERSION_CODE(2,6,15))
static void mxpcdrv_receive_chars(struct mxpcdrv_struct *info,
					 int *status)
{
	struct tty_struct *	tty = info->tty;
	unsigned char		ch, gdl;
	int			ignored = 0;
	int			cnt = 0;
	int			count;
	int 			recv_room;
	int			max = 256;
	unsigned long		flags;

        count = 0;

	// following add by Victor Yu. 09-02-2002
	if ( info->IsMoxaMustChipFlag != MOXA_OTHER_UART) {

		if ( *status & UART_LSR_SPECIAL ) {
			goto intr_old;
		}

		// following add by Victor Yu. 02-11-2004
		//if ( info->IsMoxaMustChipFlag == MOXA_MUST_MU860_HWID &&
		//     (*status & MOXA_MUST_LSR_RERR) )
		//	goto intr_old;
		// above add by Victor Yu. 02-14-2004
		if(*status & MOXA_MUST_LSR_RERR)
			goto intr_old;

		gdl = inb(info->base+MOXA_MUST_GDL_REGISTER);
		
		if ( info->IsMoxaMustChipFlag == MOXA_MUST_HARDWARE_ID ) // add by Victor Yu. 02-11-2004
			gdl &= MOXA_MUST_GDL_MASK;
		
		while ( gdl-- ) {
		    	ch = inb(info->base + UART_RX);
			if(cnt>=TTY_FLIPBUF_SIZE)
				continue;
	    		count++;
			tty_insert_flip_char(tty, ch, 0);
			cnt++;
		}
		goto end_intr;
	}
intr_old:
	// above add by Victor Yu. 09-02-2002

	do {
	    if(max-- <0)
 		break;

	    ch = inb(info->base + UART_RX);
	    if(cnt>=TTY_FLIPBUF_SIZE)
		continue;
	    // following add by Victor Yu. 09-02-2002
	    if ( info->IsMoxaMustChipFlag && (*status&UART_LSR_OE) /*&& !(*status&UART_LSR_DR)*/ )
		outb(0x23, info->base+UART_FCR);
	    *status &= info->read_status_mask;
	    // above add by Victor Yu. 09-02-2002
	    if ( *status & info->ignore_status_mask ) {
		if ( ++ignored > 100 )
		    break;
	    } else {
		count++;
		if ( *status & UART_LSR_SPECIAL ) {
		    if ( *status & UART_LSR_BI ) {
			flags = TTY_BREAK;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.brk++;
#endif
			if ( info->flags & ASYNC_SAK )
			    do_SAK(tty);
		    } else if ( *status & UART_LSR_PE ) {
			flags = TTY_PARITY;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.parity++;
#endif
		    } else if ( *status & UART_LSR_FE ) {
			flags = TTY_FRAME;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.frame++;
#endif
		    } else if ( *status & UART_LSR_OE ) {
			flags = TTY_OVERRUN;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.overrun++;
#endif
		    } else
			flags = TTY_BREAK;

		} else
			flags = 0;
		tty_insert_flip_char(tty, ch, flags);
		cnt++;

	    }
	
	    // following add by Victor Yu. 09-02-2002
	    if ( info->IsMoxaMustChipFlag )
		break;
	    // above add by Victor Yu. 09-02-2002

	    /* mask by Victor Yu. 09-02-2002
	    *status = inb(info->base + UART_LSR) & info->read_status_mask;
	    */
	    // following add by Victor Yu. 09-02-2002
	    *status = inb(info->base+UART_LSR);
	    // above add by Victor Yu. 09-02-2002
	} while ( *status & UART_LSR_DR );

end_intr:	// add by Victor Yu. 09-02-2002

	mxvar_log.rxcnt[info->port] += cnt;

	recv_room = tty->receive_room;
	if(recv_room < (128*2)){
		set_bit(TTY_THROTTLED, &tty->flags);
		mxpcdrv_stoprx(tty);
	}
	tty_flip_buffer_push(tty);
}
#else
static void mxpcdrv_receive_chars(struct mxpcdrv_struct *info,
					 int *status)
{
	struct tty_struct *	tty = info->tty;
	unsigned char		ch, gdl;
	int			ignored = 0;
	int			cnt = 0;
	int			count;
	int 			recv_room;
	unsigned char           *cp;
	char                    *fp;
	int			max = 256;


	//MX_LOCK(&info->slock);

#if 0
	recv_room = tty->ldisc.receive_room(tty);
	if((recv_room == 0) && (!info->ldisc_stop_rx)){
		//mxser_throttle(tty);
       	mxpcdrv_stoprx(tty);
		//return;
	}
#endif
		
        cp = tty->flip.char_buf;
        fp = tty->flip.flag_buf;
        count = 0;

	// following add by Victor Yu. 09-02-2002
	if ( info->IsMoxaMustChipFlag ) {
		if ( *status & UART_LSR_SPECIAL ) {
			goto intr_old;
		}

		// following add by Victor Yu. 02-11-2004
		//if ( info->IsMoxaMustChipFlag == MOXA_MUST_MU860_HWID &&
		//     (*status & MOXA_MUST_LSR_RERR) )
		//	goto intr_old;
		// above add by Victor Yu. 02-14-2004
		if(*status & MOXA_MUST_LSR_RERR)
			goto intr_old;

		gdl = inb(info->base+MOXA_MUST_GDL_REGISTER);
		//gdl &= MOXA_MUST_GDL_MASK;
		
		if ( info->IsMoxaMustChipFlag == MOXA_MUST_HARDWARE_ID ) // add by Victor Yu. 02-11-2004
			gdl &= MOXA_MUST_GDL_MASK;

#if 0		
		if(gdl >= recv_room) {
			if(!info->ldisc_stop_rx){
				//mxser_throttle(tty);
        			mxpcdrv_stoprx(tty);
			}
			//return;
		}		
#endif
		
		while ( gdl-- ) {
	    	ch = inb(info->base + UART_RX);
			/*if ( tty->flip.count >= TTY_FLIPBUF_SIZE )
		    		break;
			tty->flip.count++;
			*tty->flip.char_buf_ptr++ = ch;*/
			if(cnt>=TTY_FLIPBUF_SIZE)
				continue;
				
    		count++;
			*cp++ = ch;
	    	*fp++ = 0;
			cnt++;
		}
		goto end_intr;
	}
intr_old:
	// above add by Victor Yu. 09-02-2002

	do {
	    if(max-- <0)
 		break;

	    ch = inb(info->base + UART_RX);
			if(cnt>=TTY_FLIPBUF_SIZE)
				continue;

	    // following add by Victor Yu. 09-02-2002
	    if ( info->IsMoxaMustChipFlag && (*status&UART_LSR_OE) /*&& !(*status&UART_LSR_DR)*/ )
		outb(0x23, info->base+UART_FCR);
	    *status &= info->read_status_mask;
	    // above add by Victor Yu. 09-02-2002
	    if ( *status & info->ignore_status_mask ) {
		if ( ++ignored > 100 )
		    break;
	    } else {
		/*if ( tty->flip.count >= TTY_FLIPBUF_SIZE )
		    break;
		tty->flip.count++;
		*/
		count++;
		if ( *status & UART_LSR_SPECIAL ) {
		    if ( *status & UART_LSR_BI ) {
			*fp++ = TTY_BREAK;
/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.brk++;
#endif

/* */
			if ( info->flags & ASYNC_SAK )
			    do_SAK(tty);
		    } else if ( *status & UART_LSR_PE ) {
			*fp++ = TTY_PARITY;
/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.parity++;
#endif
/* */
		    } else if ( *status & UART_LSR_FE ) {
			*fp++ = TTY_FRAME;
/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.frame++;
#endif
/* */
		    } else if ( *status & UART_LSR_OE ) {
			*fp++ = TTY_OVERRUN;
/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
			info->icount.overrun++;
#endif
/* */
		    } else
			*fp++ = 0;
		} else
		    *fp++ = 0;
		    *cp++ = ch;
		cnt++;
#if 0
		if(cnt>=recv_room){
			if(!info->ldisc_stop_rx){
				//mxser_throttle(tty);
        			mxpcdrv_stoprx(tty);
			}
			break;
		}
#endif
		}
	
	    // following add by Victor Yu. 09-02-2002
	    if ( info->IsMoxaMustChipFlag )
		break;
	    // above add by Victor Yu. 09-02-2002

	    /* mask by Victor Yu. 09-02-2002
	    *status = inb(info->base + UART_LSR) & info->read_status_mask;
	    */
	    // following add by Victor Yu. 09-02-2002
	    *status = inb(info->base+UART_LSR);
	    // above add by Victor Yu. 09-02-2002
	} while ( *status & UART_LSR_DR );

end_intr:	// add by Victor Yu. 09-02-2002

	mxvar_log.rxcnt[info->port] += cnt;
/* added by casper 1/11/2000 */
#if 0
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	info->icount.rx += cnt;

	/* add by Victor Yu. 05-14-2002 following */
	clear_bit(TTY_DONT_FLIP, &tty->flags);
	/* add by Victor Yu. 05-14-2002 above */

	tty_flip_buffer_push(tty);
#else
	queue_task_irq_off(&tty->flip.tqueue, &tq_timer);
#endif
#endif
#if 1 
	tty->ldisc.receive_buf(tty, tty->flip.char_buf, tty->flip.flag_buf, count);
	//MX_UNLOCK(&info->slock);
	recv_room = tty->ldisc.receive_room(tty);
	if(recv_room < (128*2)){

		set_bit(TTY_THROTTLED, &tty->flags);
		mxpcdrv_stoprx(tty);
	}
#endif
/* */
}
#endif//_SCREEN_INFO_H

static void mxpcdrv_transmit_chars(struct mxpcdrv_struct *info)
{
	int	count, cnt;

	if ( !info)
		return;
	if ( info->x_char ) {
	    outb(info->x_char, info->base + UART_TX);
	    info->x_char = 0;
	    mxvar_log.txcnt[info->port]++;
/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	    info->icount.tx++;
#endif
/* */
	    return;
	}

if ( info->xmit_buf == 0 )
	return;

	if(info->xmit_cnt==0){
		if ( info->xmit_cnt < WAKEUP_CHARS ) {
			set_bit(MXPCDRV_EVENT_TXLOW,&info->event);
			MXQ_TASK();
		}
		return;
	}
#if 1
	if (/*(info->xmit_cnt <= 0) ||*/ info->tty->stopped ||
	    (info->tty->hw_stopped && (info->type != PORT_16550A) &&(!info->IsMoxaMustChipFlag))) {
		info->IER &= ~UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
		return;
	}
#endif

	cnt = info->xmit_cnt;
	count = info->xmit_fifo_size;
	do {
	    outb(info->xmit_buf[info->xmit_tail++], info->base + UART_TX);
	    info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE - 1);
	    if ( --info->xmit_cnt <= 0 )
		break;
	} while ( --count > 0 );
	mxvar_log.txcnt[info->port] += (cnt - info->xmit_cnt);

/* added by casper 1/11/2000 */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        info->icount.tx += (cnt - info->xmit_cnt);
#endif
/* */

	if ( info->xmit_cnt < WAKEUP_CHARS ) {
		set_bit(MXPCDRV_EVENT_TXLOW,&info->event);
		MXQ_TASK();
	}
#if 0
	if (info->xmit_cnt <= 0) {
		info->IER &= ~UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
	}
#endif

}

static void mxpcdrv_check_modem_status(struct mxpcdrv_struct *info,
					      int status)
{
	/* update input line counters */
	if ( status & UART_MSR_TERI )
	    info->icount.rng++;
	if ( status & UART_MSR_DDSR )
	    info->icount.dsr++;
	if ( status & UART_MSR_DDCD )
	    info->icount.dcd++;
	if ( status & UART_MSR_DCTS )
	    info->icount.cts++;
	wake_up_interruptible(&info->delta_msr_wait);

	if ( (info->flags & ASYNC_CHECK_CD) && (status & UART_MSR_DDCD) ) {
	    if ( status & UART_MSR_DCD )
		wake_up_interruptible(&info->open_wait);
	    else if ( !((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		      (info->flags & ASYNC_CALLOUT_NOHUP)) )

	    set_bit(MXPCDRV_EVENT_HANGUP,&info->event);
		MXQ_TASK();
	}
	if ( info->flags & ASYNC_CTS_FLOW ) {
	    if ( info->tty->hw_stopped ) {
		if (status & UART_MSR_CTS ){  
		    info->tty->hw_stopped = 0;
#if 0
		    if ((info->type != PORT_16550A)&& (!info->IsMoxaMustChipFlag)) {
					info->IER &= ~UART_IER_THRI;
					outb(info->IER, info->base + UART_IER);
					info->IER |= UART_IER_THRI;
					outb(info->IER, info->base + UART_IER);
		    }
#endif
	       	    set_bit(MXPCDRV_EVENT_TXLOW,&info->event);
				MXQ_TASK();
                }   
	    } else {
		if ( !(status & UART_MSR_CTS) ){
		    info->tty->hw_stopped = 1;
#if 0
		    if ((info->type != PORT_16550A)&& (!info->IsMoxaMustChipFlag)) {
			info->IER &= ~UART_IER_THRI;
			outb(info->IER, info->base + UART_IER);
		    }
#endif
		}
	    }
	}
}

static int mxpcdrv_block_til_ready(struct tty_struct *tty, struct file * filp,
				 struct mxpcdrv_struct *info)
{
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
	DECLARE_WAITQUEUE(wait, current); 
#else
	struct wait_queue	wait = { current, NULL };
#endif
	int			retval;
	int			do_clocal = 0;
	MX_LOCK_INIT();

#if 0
	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if ( tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING) ) {
	    if ( info->flags & ASYNC_CLOSING )
		interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
	    if ( info->flags & ASYNC_HUP_NOTIFY )
		return(-EAGAIN);
	    else
		return(-ERESTARTSYS);
#else
	    return(-EAGAIN);
#endif
	}
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,4,0))
	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if ( MX_TTY_DRV(subtype) == SERIAL_TYPE_CALLOUT ) {
	    if ( info->flags & ASYNC_NORMAL_ACTIVE )
		return(-EBUSY);
	    if ( (info->flags & ASYNC_CALLOUT_ACTIVE) &&
		 (info->flags & ASYNC_SESSION_LOCKOUT) &&
		 (info->session != MX_SESSION()) )
		return(-EBUSY);
	    if ( (info->flags & ASYNC_CALLOUT_ACTIVE) &&
		 (info->flags & ASYNC_PGRP_LOCKOUT) &&
		 (info->pgrp != MX_CGRP()) )
		return(-EBUSY);
	    info->flags |= ASYNC_CALLOUT_ACTIVE;
	    return(0);
	}
#endif

	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ( (filp->f_flags & O_NONBLOCK) ||
	     (tty->flags & (1 << TTY_IO_ERROR)) ) {
	    if ( info->flags & ASYNC_CALLOUT_ACTIVE )
		return(-EBUSY);
	    info->flags |= ASYNC_NORMAL_ACTIVE;
	    return(0);
	}

	if ( info->flags & ASYNC_CALLOUT_ACTIVE ) {
	    if ( info->normal_termios.c_cflag & CLOCAL )
		do_clocal = 1;
	} else {
	    if ( tty->termios->c_cflag & CLOCAL )
		do_clocal = 1;
	}

	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
	 * mxpcdrv_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
	MX_LOCK(&info->slock);
	if ( !tty_hung_up_p(filp) )
	    info->count--;
	MX_UNLOCK(&info->slock);
	info->blocked_open++;
	while ( 1 ) {
	    MX_LOCK(&info->slock);
	    if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) )
		outb(inb(info->base + UART_MCR) | UART_MCR_DTR | UART_MCR_RTS,
		     info->base + UART_MCR);
	    MX_UNLOCK(&info->slock);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
			set_current_state(TASK_INTERRUPTIBLE);
#else
			current->state = TASK_INTERRUPTIBLE;
#endif
	    if ( tty_hung_up_p(filp) || !(info->flags & ASYNC_INITIALIZED) ) {
#ifdef SERIAL_DO_RESTART
		if ( info->flags & ASYNC_HUP_NOTIFY )
		    retval = -EAGAIN;
		else
		    retval = -ERESTARTSYS;
#else
		retval = -EAGAIN;
#endif
		break;
	    }
	    if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		 !(info->flags & ASYNC_CLOSING) &&
		 (do_clocal || (inb(info->base + UART_MSR) & UART_MSR_DCD)) )
		break;
	    if ( signal_pending(current) ) {
		retval = -ERESTARTSYS;
		break;
	    }
	    schedule();
	}
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
		set_current_state(TASK_RUNNING);
#else
		current->state = TASK_RUNNING;
#endif
	remove_wait_queue(&info->open_wait, &wait);
	if ( !tty_hung_up_p(filp) )
	    info->count++;
	info->blocked_open--;
	if ( retval )
	    return(retval);
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return(0);
}

static int mxpcdrv_startup(struct mxpcdrv_struct * info)
{
	unsigned long	page;
	MX_LOCK_INIT();

	page = GET_FPAGE(GFP_KERNEL);
	if ( !page )
	    return(-ENOMEM);

	MX_LOCK(&info->slock);
	if ( info->flags & ASYNC_INITIALIZED ) {
	    free_page(page);
	    MX_UNLOCK(&info->slock);
	    return(0);
	}

	if ( !info->base || !info->type ) {
	    if ( info->tty )
		set_bit(TTY_IO_ERROR, &info->tty->flags);
	    free_page(page);
	    MX_UNLOCK(&info->slock);
	    return(0);
	}
	if ( info->xmit_buf )
	    free_page(page);
	else
	    info->xmit_buf = (unsigned char *)page;

	/*
	 * Clear the FIFO buffers and disable them
	 * (they will be reenabled in mxpcdrv_change_speed())
	 */
	if ( info->xmit_fifo_size == 16 )
	    outb((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT),
		 info->base + UART_FCR);
	// following add by Victor Yu. 08-30-2002
	else if ( info->xmit_fifo_size == 64 )
	    outb((UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT|MOXA_MUST_FCR_GDA_MODE_ENABLE), info->base+UART_FCR);
	// above add by Victor Yu. 08-30-2002

	/*
	 * At this point there's no way the LSR could still be 0xFF;
	 * if it is, then bail out, because there's likely no UART
	 * here.
	 */
	if ( inb(info->base + UART_LSR) == 0xff ) {
	    MX_UNLOCK(&info->slock);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
	    if ( suser() ) {
#else
            if (capable(CAP_SYS_ADMIN)) {
#endif	    	
		if ( info->tty )
		    set_bit(TTY_IO_ERROR, &info->tty->flags);
		return(0);
	    } else
		return(-ENODEV);
	}
         
	/*
	 * Clear the interrupt registers.
	 */
	(void)inb(info->base + UART_LSR);
	(void)inb(info->base + UART_RX);
	(void)inb(info->base + UART_IIR);
	(void)inb(info->base + UART_MSR);

	/*
	 * Now, initialize the UART
	 */
	outb(UART_LCR_WLEN8, info->base + UART_LCR);	/* reset DLAB */
	info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	outb(info->MCR, info->base + UART_MCR);

	/*
	 * Finally, enable interrupts
	 */
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;

	// following add by Victor Yu. 08-30-2002
	if ( info->IsMoxaMustChipFlag )
		info->IER |= MOXA_MUST_IER_EGDAI;
	// above add by Victor Yu. 08-30-2002

	outb(info->IER, info->base + UART_IER); /* enable interrupts */

	/*
	 * And clear the interrupt registers again for luck.
	 */
	(void)inb(info->base + UART_LSR);
	(void)inb(info->base + UART_RX);
	(void)inb(info->base + UART_IIR);
	(void)inb(info->base + UART_MSR);

#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
	if ( info->tty )
	    clear_bit(TTY_IO_ERROR, &info->tty->flags);
#else
	if ( info->tty )
	    test_and_clear_bit(TTY_IO_ERROR, &info->tty->flags);
#endif
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;

	/*
	 * and set the speed of the serial port
	 */
	MX_UNLOCK(&info->slock);
	mxpcdrv_change_speed(info, 0);

	info->flags |= ASYNC_INITIALIZED;
	return(0);
}

/*
 * This routine will shutdown a serial port; interrupts maybe disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void mxpcdrv_shutdown(struct mxpcdrv_struct * info)
{
	MX_LOCK_INIT();

	if ( !(info->flags & ASYNC_INITIALIZED) )
	    return;

    MX_LOCK(&info->slock);
	/* Disable interrupts */

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
	wake_up_interruptible(&info->delta_msr_wait);

	/*
	 * Free the IRQ, if necessary
	 */
	if ( info->xmit_buf ) {
	    free_page((unsigned long)info->xmit_buf);
	    info->xmit_buf = 0;
	}

	info->IER = 0;
	outb(0x00, info->base + UART_IER);

	if ( !info->tty || (info->tty->termios->c_cflag & HUPCL) )
	    info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);
	outb(info->MCR, info->base + UART_MCR);

	/* clear Rx/Tx FIFO's */
	// following add by Victor Yu. 08-30-2002
	if ( info->IsMoxaMustChipFlag )
		outb((UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT|MOXA_MUST_FCR_GDA_MODE_ENABLE), info->base + UART_FCR);
	else
	// above add by Victor Yu. 08-30-2002
		outb((UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT), info->base + UART_FCR);

	/* read data port to reset things */
	(void)inb(info->base + UART_RX);

	if ( info->tty )
	    set_bit(TTY_IO_ERROR, &info->tty->flags);

	info->flags &= ~ASYNC_INITIALIZED;

	// following add by Victor Yu. 09-23-2002
	if ( info->IsMoxaMustChipFlag ) {
		SET_MOXA_MUST_NO_SOFTWARE_FLOW_CONTROL(info->base);
	}
	// above add by Victor Yu. 09-23-2002

	MX_UNLOCK(&info->slock);
}

/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a serial port.
 */
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
static int mxpcdrv_change_speed(struct mxpcdrv_struct *info,
                              struct ktermios *old_termios)
#else
static int mxpcdrv_change_speed(struct mxpcdrv_struct *info,
                              struct termios *old_termios)
#endif
{
	int		quot = 0;
	unsigned	cflag, cval, fcr;
	int		i;
        int             ret = 0;
	MX_LOCK_INIT();

	if ( !info->tty || !info->tty->termios )
	    return ret;
	cflag = info->tty->termios->c_cflag;
	if ( !(info->base) )
	    return ret;
           
#ifndef B921600
#define B921600 (B460800 +1)
#endif 
	switch( cflag & (CBAUD | CBAUDEX) ){
        case B921600 : i = 20; break;
        case B460800 : i = 19; break;
        case B230400 : i = 18; break;
        case B115200 : i = 17; break;
        case B57600 : i = 16; break;
        case B38400 : i = 15; break;
        case B19200 : i = 14; break;
        case B9600 : i = 13; break;
        case B4800 : i = 12; break;
        case B2400 : i = 11; break;
        case B1800 : i = 10; break;
        case B1200 : i = 9; break;
        case B600 : i = 8; break;
        case B300 : i = 7; break;
        case B200 : i = 6; break;
        case B150 : i = 5; break;
        case B134 : i = 4; break;
        case B110 : i = 3; break;
        case B75 : i = 2; break;
        case B50 : i = 1; break;
        default:   i = 0; break;
	}
	
	if ( i == 15 ) {
	    if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI )
		i = 16; 					/* 57600 bps */
	    if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI )
		i = 17; 					/* 115200 bps */
                
#ifdef ASYNC_SPD_SHI
	    if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
                i = 18;
#endif

#ifdef ASYNC_SPD_WARP
            if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
                i = 19;
#endif
	}
	if ( mxvar_baud_table[i] == 134 ) {
	    quot = (2 * info->baud_base / 269);
	} else if ( mxvar_baud_table[i] ) {
	    quot = info->baud_base / mxvar_baud_table[i];
	    // following add by Victor Yu. 09-04-2002
	    if ( mxvar_baud_table[i] > info->MaxCanSetBaudRate )
		quot = 0;
	    // add by Victor Yu. 09-04-2002
       	    if (!quot && old_termios) {
                /* re-calculate */
        	info->tty->termios->c_cflag &= ~CBAUD;
        	info->tty->termios->c_cflag |= (old_termios->c_cflag & CBAUD);
		switch( info->tty->termios->c_cflag & (CBAUD | CBAUDEX) ){
	        case B921600 : i = 20; break;
        	case B460800 : i = 19; break;
	        case B230400 : i = 18; break;
        	case B115200 : i = 17; break;
	        case B57600 : i = 16; break;
        	case B38400 : i = 15; break;
	        case B19200 : i = 14; break;
        	case B9600 : i = 13; break;
	        case B4800 : i = 12; break;
        	case B2400 : i = 11; break;
	        case B1800 : i = 10; break;
        	case B1200 : i = 9; break;
	        case B600 : i = 8; break;
        	case B300 : i = 7; break;
	        case B200 : i = 6; break;
        	case B150 : i = 5; break;
	        case B134 : i = 4; break;
        	case B110 : i = 3; break;
	        case B75 : i = 2; break;
        	case B50 : i = 1; break;
	        default:   i = 0; break;
		}
        	if ( i == 15 ) {
        	    if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI )
        	        i = 16; 					/* 57600 bps */
        	    if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI )
        	        i = 17; 					/* 115200 bps */
#ifdef ASYNC_SPD_SHI
	            if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
                        i = 18;
#endif
#ifdef ASYNC_SPD_WARP
                    if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
                        i = 19;
#endif
        	}
        	if ( mxvar_baud_table[i] == 134 ) {
        	    quot = (2 * info->baud_base / 269);
                } else if ( mxvar_baud_table[i] ) {
        	    quot = info->baud_base / mxvar_baud_table[i];
                    if(quot==0)
                        quot = 1;
		    // following add by Victor Yu. 09-04-2002
		    if ( mxvar_baud_table[i] > info->MaxCanSetBaudRate )
			quot = 0;
		    // above add by Victor Yu. 09-04-2002
        	} else {
        	    quot = 0;
        	}
            }else if(quot==0)
                quot = 1;
	} else {
	    quot = 0;
	}

	info->timeout = (int)((unsigned int)(info->xmit_fifo_size*HZ*10*quot) / (unsigned int)info->baud_base);
	info->timeout += HZ/50;		/* Add .02 seconds of slop */

	if ( quot ) {
	    info->MCR |= UART_MCR_DTR;
	    MX_LOCK(&info->slock);
	    outb(info->MCR, info->base + UART_MCR);
	    MX_UNLOCK(&info->slock);
	} else {
	    info->MCR &= ~UART_MCR_DTR;
	    MX_LOCK(&info->slock);
	    outb(info->MCR, info->base + UART_MCR);
	    MX_UNLOCK(&info->slock);
	    return ret;
	}
	/* byte size and parity */
	switch ( cflag & CSIZE ) {
	case CS5: cval = 0x00; break;
	case CS6: cval = 0x01; break;
	case CS7: cval = 0x02; break;
	case CS8: cval = 0x03; break;
	default:  cval = 0x00; break;	/* too keep GCC shut... */
	}
	if ( cflag & CSTOPB )
	    cval |= 0x04;
	if ( cflag & PARENB )
	    cval |= UART_LCR_PARITY;
#ifndef CMSPAR
#define	CMSPAR 010000000000
#endif
	if ( !(cflag & PARODD) ){
	    cval |= UART_LCR_EPAR;
	}
	if ( cflag & CMSPAR )
	    cval |= UART_LCR_SPAR;

	if ( (info->type == PORT_8250) || (info->type == PORT_16450) ) {
	    fcr = 0;
	} else {
	    fcr = UART_FCR_ENABLE_FIFO;
	    // following add by Victor Yu. 08-30-2002
	    if ( info->IsMoxaMustChipFlag ) {
		fcr |= MOXA_MUST_FCR_GDA_MODE_ENABLE;
		SET_MOXA_MUST_FIFO_VALUE(info);
	    } else {
	    // above add by Victor Yu. 08-30-2002

	    	switch ( info->rx_trigger ) {
	    	case 1:  fcr |= UART_FCR_TRIGGER_1; break;
	    	case 4:  fcr |= UART_FCR_TRIGGER_4; break;
	    	case 8:  fcr |= UART_FCR_TRIGGER_8; break;
	    	default: fcr |= UART_FCR_TRIGGER_14;break;
	    	}
	    }
	}

	/* CTS flow control flag and modem status interrupts */
	info->IER &= ~UART_IER_MSI;
	info->MCR &= ~UART_MCR_AFE;
	if ( cflag & CRTSCTS ) {
	    info->flags |= ASYNC_CTS_FLOW;
	    info->IER |= UART_IER_MSI;
	    if (( info->type == PORT_16550A ) || (info->IsMoxaMustChipFlag)){
		info->MCR |= UART_MCR_AFE;
		}else {
			unsigned char status;
	//MX_LOCK(&info->slock);
			status = inb(info->base + UART_MSR);
	//MX_UNLOCK(&info->slock);
			if (info->tty->hw_stopped) {
				if (status & UART_MSR_CTS) {
					info->tty->hw_stopped = 0;
#if 0
					if ((info->type != PORT_16550A) && (!info->IsMoxaMustChipFlag)){
						info->IER &= ~UART_IER_THRI;
						outb(info->IER, info->base + UART_IER);
						info->IER |= UART_IER_THRI;
						outb(info->IER, info->base + UART_IER);
					}
#endif
					set_bit(MXPCDRV_EVENT_TXLOW, &info->event);
					MXQ_TASK();
				}
			} else {
				if (!(status & UART_MSR_CTS)) {
					info->tty->hw_stopped = 1;
#if 0
					if ((info->type != PORT_16550A) && (!info->IsMoxaMustChipFlag)){
						info->IER &= ~UART_IER_THRI;
						outb(info->IER, info->base + UART_IER);
					}
#endif
				}
			}
	        }
	} else {
	    info->flags &= ~ASYNC_CTS_FLOW;
	}
	outb(info->MCR, info->base + UART_MCR);
	if ( cflag & CLOCAL ){
	    info->flags &= ~ASYNC_CHECK_CD;
	}else {
	    info->flags |= ASYNC_CHECK_CD;
	    info->IER |= UART_IER_MSI;
	}
	outb(info->IER, info->base + UART_IER);

	/*
	 * Set up parity check flag
	 */
	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
	if ( I_INPCK(info->tty) )
	    info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if ( I_BRKINT(info->tty) || I_PARMRK(info->tty) )
	    info->read_status_mask |= UART_LSR_BI;

	info->ignore_status_mask = 0;
#if 0
	/* This should be safe, but for some broken bits of hardware... */
	if ( I_IGNPAR(info->tty) ) {
	    info->ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
	    info->read_status_mask |= UART_LSR_PE | UART_LSR_FE;
	}
#endif
	if ( I_IGNBRK(info->tty) ) {
	    info->ignore_status_mask |= UART_LSR_BI;
	    info->read_status_mask |= UART_LSR_BI;
	    /*
	     * If we're ignore parity and break indicators, ignore
	     * overruns too.  (For real raw support).
	     */
	    if ( I_IGNPAR(info->tty) ) {
		info->ignore_status_mask |= UART_LSR_OE|UART_LSR_PE|UART_LSR_FE;
		info->read_status_mask |= UART_LSR_OE|UART_LSR_PE|UART_LSR_FE;
	    }
	}
	
	// following add by Victor Yu. 09-02-2002
	if ( info->IsMoxaMustChipFlag ) {
	    MX_LOCK(&info->slock);
		SET_MOXA_MUST_XON1_VALUE(info->base, START_CHAR(info->tty));
		SET_MOXA_MUST_XOFF1_VALUE(info->base, STOP_CHAR(info->tty));
		if ( I_IXON(info->tty) ) {
			ENABLE_MOXA_MUST_RX_SOFTWARE_FLOW_CONTROL(info->base);
		} else {
			DISABLE_MOXA_MUST_RX_SOFTWARE_FLOW_CONTROL(info->base);
		}
		if ( I_IXOFF(info->tty) ) {
			ENABLE_MOXA_MUST_TX_SOFTWARE_FLOW_CONTROL(info->base);
		} else {
			DISABLE_MOXA_MUST_TX_SOFTWARE_FLOW_CONTROL(info->base);
		}
#if 0
		if ( I_IXANY(info->tty) ) {
			info->MCR |= MOXA_MUST_MCR_XON_ANY;
			ENABLE_MOXA_MUST_XON_ANY_FLOW_CONTROL(info->base);
		} else {
			info->MCR &= ~MOXA_MUST_MCR_XON_ANY;
			DISABLE_MOXA_MUST_XON_ANY_FLOW_CONTROL(info->base);
		}
#endif
		MX_UNLOCK(&info->slock);
	}
	// above add by Victor Yu. 09-02-2002

	MX_LOCK(&info->slock);	
	outb(cval | UART_LCR_DLAB, info->base + UART_LCR);  /* set DLAB */
	outb(quot & 0xff, info->base + UART_DLL);	    /* LS of divisor */
	outb(quot >> 8, info->base + UART_DLM); 	    /* MS of divisor */
	outb(cval, info->base + UART_LCR);		    /* reset DLAB */
	outb(fcr, info->base + UART_FCR);		    /* set fcr */
	MX_UNLOCK(&info->slock);

        return ret;
}

/*
 * ------------------------------------------------------------
 * friends of mxpcdrv_ioctl()
 * ------------------------------------------------------------
 */
static int mxpcdrv_get_serial_info(struct mxpcdrv_struct * info,
				 struct serial_struct * retinfo)
{
	struct serial_struct	tmp;

	if ( !retinfo )
	    return(-EFAULT);
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->type;
	tmp.line = info->port;
	tmp.port = info->base;
	tmp.irq = info->irq;
	tmp.flags = info->flags;
	tmp.baud_base = info->baud_base;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	tmp.hub6 = 0;
	if(copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
	    return -EFAULT;
	return(0);
}

static int mxpcdrv_set_serial_info(struct mxpcdrv_struct * info,
				 struct serial_struct * new_info)
{
	struct serial_struct	new_serial;
	unsigned int		flags;
	int			retval = 0;

	if ( !new_info || !info->base )
	    return(-EFAULT);
	if(copy_from_user(&new_serial, new_info, sizeof(new_serial)))
	    return -EFAULT;

	if ( (new_serial.irq != info->irq) ||
	     (new_serial.port != info->base) ||
	     (new_serial.custom_divisor != info->custom_divisor) ||
	     (new_serial.baud_base != info->baud_base) )
	    return(-EPERM);

	flags = info->flags & ASYNC_SPD_MASK;

#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
       if ( !suser() ) {
#else
       if ( !capable(CAP_SYS_ADMIN)) {
#endif	    	
	    if ( (new_serial.baud_base != info->baud_base) ||
		 (new_serial.close_delay != info->close_delay) ||
		 ((new_serial.flags & ~ASYNC_USR_MASK) !=
		 (info->flags & ~ASYNC_USR_MASK)) )
		return(-EPERM);
	    info->flags = ((info->flags & ~ASYNC_USR_MASK) |
			  (new_serial.flags & ASYNC_USR_MASK));
	} else {
	    /*
	     * OK, past this point, all the error checking has been done.
	     * At this point, we start making changes.....
	     */
	    info->flags = ((info->flags & ~ASYNC_FLAGS) |
			  (new_serial.flags & ASYNC_FLAGS));
	    info->close_delay = new_serial.close_delay * HZ/100;
	    info->closing_wait = new_serial.closing_wait * HZ/100;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
	    info->tty->low_latency = 1;
#endif
	}

/* added by casper, 3/17/2000, for mouse */
	info->type = new_serial.type;

	// following add by Victor Yu. 09-02-2002
	if ( info->IsMoxaMustChipFlag )
		info->xmit_fifo_size = 64;
	else
	// above add by Victor Yu. 09-02-2002
	if(info->type == PORT_16550A)
		info->xmit_fifo_size = 16;
	else
		info->xmit_fifo_size = 1;
/* */
	if ( info->flags & ASYNC_INITIALIZED ) {
	    if ( flags != (info->flags & ASYNC_SPD_MASK) ){
		mxpcdrv_change_speed(info,0);
	    }
	} else{
	    retval = mxpcdrv_startup(info);
	}
	return(retval);
}

/*
 * mxpcdrv_get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 *	    is emptied.  On bus types like RS485, the transmitter must
 *	    release the bus after transmitting. This must be done when
 *	    the transmit shift register is empty, not be done when the
 *	    transmit holding register is empty.  This functionality
 *	    allows an RS485 driver to be written in user space.
 */
static int mxpcdrv_get_lsr_info(struct mxpcdrv_struct * info, unsigned int *value)
{
	unsigned char	status;
	unsigned int	result;
	MX_LOCK_INIT();

	MX_LOCK(&info->slock);
	status = inb(info->base + UART_LSR);
	MX_UNLOCK(&info->slock);
	result = ((status & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);
	put_to_user(result, value);
	return(0);
}

/*
 * This routine sends a break character out the serial port.
 */
static void mxpcdrv_send_break(struct mxpcdrv_struct * info, int duration)
{
	MX_LOCK_INIT();

	if ( !info->base )
	    return;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,3,0))
	set_current_state(TASK_INTERRUPTIBLE);
#else
	current->state = TASK_INTERRUPTIBLE;
#endif
	MX_LOCK(&info->slock);	
	outb(inb(info->base + UART_LCR) | UART_LCR_SBC, info->base + UART_LCR);
	schedule_timeout(duration);
	outb(inb(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);
	MX_UNLOCK(&info->slock);
}

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static int mxpcdrv_tiocmget(struct tty_struct *tty, struct file *file)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *) tty->driver_data;
	unsigned char control, status;
	MX_LOCK_INIT();


	if (PORTNO(tty) == MXPCDRV_PORTS)
		return (-ENOIOCTLCMD);
	if (tty->flags & (1 << TTY_IO_ERROR))
		return (-EIO);

	control = info->MCR;
	
	MX_LOCK(&info->slock);
	status = inb(info->base + UART_MSR);
	if (status & UART_MSR_ANY_DELTA)
		mxpcdrv_check_modem_status(info, status);
	MX_UNLOCK(&info->slock);
	return ((control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
	    ((control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
	    ((status & UART_MSR_DCD) ? TIOCM_CAR : 0) |
	    ((status & UART_MSR_RI) ? TIOCM_RNG : 0) |
	    ((status & UART_MSR_DSR) ? TIOCM_DSR : 0) |
	    ((status & UART_MSR_CTS) ? TIOCM_CTS : 0);
}

static int mxpcdrv_tiocmset(struct tty_struct *tty, struct file *file,
			  unsigned int set, unsigned int clear)
{
	struct mxpcdrv_struct *info = (struct mxpcdrv_struct *) tty->driver_data;
	MX_LOCK_INIT();


	if (PORTNO(tty) == MXPCDRV_PORTS)
		return (-ENOIOCTLCMD);
	if (tty->flags & (1 << TTY_IO_ERROR))
		return (-EIO);

	MX_LOCK(&info->slock);

	if (set & TIOCM_RTS)
		info->MCR |= UART_MCR_RTS;
	if (set & TIOCM_DTR)
		info->MCR |= UART_MCR_DTR;

	if (clear & TIOCM_RTS)
		info->MCR &= ~UART_MCR_RTS;
	if (clear & TIOCM_DTR)
		info->MCR &= ~UART_MCR_DTR;

	outb(info->MCR, info->base + UART_MCR);
	MX_UNLOCK(&info->slock);
	return (0);
}

#else
static int mxpcdrv_get_modem_info(struct mxpcdrv_struct * info,
				unsigned int *value)
{
	unsigned char	control, status;
	unsigned int	result;
	MX_LOCK_INIT();

	MX_LOCK(&info->slock);
	control = info->MCR;
	status = inb(info->base + UART_MSR);
	MX_UNLOCK(&info->slock);
	
	if ( status & UART_MSR_ANY_DELTA )
	    mxpcdrv_check_modem_status(info, status);
	result = ((control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
		 ((control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
		 ((status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
		 ((status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
		 ((status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
		 ((status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
	put_to_user(result, value);
	return(0);
}

static int mxpcdrv_set_modem_info(struct mxpcdrv_struct * info, unsigned int cmd,
				unsigned int *value)
{
	int		error;
	unsigned int	arg;
	MX_LOCK_INIT();

	error = MX_ACCESS_CHK(VERIFY_READ, value, sizeof(int));
	if ( MX_ERR(error) )
	    return(error);
	get_from_user(arg,value);
	switch ( cmd ) {
	case TIOCMBIS:
	    if ( arg & TIOCM_RTS )
		info->MCR |= UART_MCR_RTS;
	    if ( arg & TIOCM_DTR )
		info->MCR |= UART_MCR_DTR;
	    break;
	case TIOCMBIC:
	    if ( arg & TIOCM_RTS )
		info->MCR &= ~UART_MCR_RTS;
	    if ( arg & TIOCM_DTR )
		info->MCR &= ~UART_MCR_DTR;
	    break;
	case TIOCMSET:
	    info->MCR = ((info->MCR & ~(UART_MCR_RTS | UART_MCR_DTR)) |
			((arg & TIOCM_RTS) ? UART_MCR_RTS : 0) |
			((arg & TIOCM_DTR) ? UART_MCR_DTR : 0));
	    break;
	default:
	    return(-EINVAL);
	}
	MX_LOCK(&info->slock);
	outb(info->MCR, info->base + UART_MCR);
	MX_UNLOCK(&info->slock);
	return(0);
}
#endif

static void	mxpcdrv_read_register(int,unsigned char *);

static int mxpcdrv_get_ISA_conf(int cap,int vector, 
		struct mxpcdrv_hwconf *hwconf,int irq)
{
	int		id, i;
	unsigned char	scratch, scratch2;
	unsigned char	value[8];

	if ( !vector )
	    return(MXPCDRV_ERR_VECTOR);

	if ( !irq ) 
	    return(MXPCDRV_ERR_IRQ);

	mxpcdrv_read_register(vector,value);

	id = value[2];
	if (id == CA104_ID){
	    hwconf->board_type = MXPCDRV_BOARD_CA104;
	    hwconf->ports = 4;
	}else if (id == CA132_ID){
	    hwconf->board_type = MXPCDRV_BOARD_CA132;
	    hwconf->ports = 2;
	}else if (id == CA132I_ID){
	    hwconf->board_type = MXPCDRV_BOARD_CA132I;
	    hwconf->ports = 2;
  }else if(id == CA108_ID){
	    hwconf->board_type = MXPCDRV_BOARD_CA108;
	    hwconf->ports = 8;
  }else if(id == CA114_ID){
	    hwconf->board_type = MXPCDRV_BOARD_CA114;
	    hwconf->ports = 4;
  }else if(id == CA134I_ID){
	    hwconf->board_type = MXPCDRV_BOARD_CA134I;
	    hwconf->ports = 4;
  }else{
	    return(0);
	}

	hwconf->irq = (int)irq;
	
	for ( i=0; i<hwconf->ports; i++ ){
		hwconf->ioaddr[i] = (int)(cap + (8 * i));
	}
	
	hwconf->vector = vector;
	if ( (id == CA104_ID) ||
			 (id == CA114_ID) ||
			 (id == CA134I_ID))
	    hwconf->vector_mask = 0x0F;
	else if(id == CA108_ID)
			hwconf->vector_mask = 0xFF;
	else 
	    hwconf->vector_mask = 0x03;

	hwconf->IsMoxaMustChipFlag = CheckIsMoxaMust(hwconf->ioaddr[0]);
	for ( i=0 ; i<hwconf->ports ; i++ ){
	    hwconf->baud_base[i] = 921600;
/*
	    if(hwconf->board_type < MXPCDRV_BOARD_INDEX)
	    		hwconf->MaxCanSetBaudRate[i] = 230400;
	    else
*/
	    		hwconf->MaxCanSetBaudRate[i] = 921600;
	}

	scratch2 = inb(cap + UART_LCR) & (~UART_LCR_DLAB);
	outb(scratch2 | UART_LCR_DLAB, cap + UART_LCR);
	outb(0, cap + UART_EFR);	/* EFR is the same as FCR */
	outb(scratch2, cap + UART_LCR);
	outb(UART_FCR_ENABLE_FIFO, cap + UART_FCR);
	scratch = inb(cap + UART_IIR);

	if ( scratch & 0xC0 )
	    hwconf->uart_type = PORT_16550A;
	else
	    hwconf->uart_type = PORT_16450;
	
	return(hwconf->ports);
}

static void mxpcdrv_read_register(int vect, unsigned char *buf)
{
	int i;

	for(i=0;i<8;i++){
		buf[i] = inb((vect+i));
	}
	return;
}

static void mxpcdrv_set_special_baudrate(struct mxpcdrv_struct * info, int baud)
{
	unsigned int	eno;
	unsigned long	n;
	unsigned char	val;
	int 		speed;

	if ( baud < 50 || baud > 921600)
		return ;
		
	speed = baud;
	info->cur_baud = baud;
	val = inb(info->base + UART_LCR);
	eno = 0;

	if ( (baud = UART_CLK_DIV / speed) == 0 ) {
		baud = 1;
	}
	else {
		n = (unsigned long)(UART_CLK_DIV % speed);
		if ( n ) {
			n <<= 3;       /* n = n * 8 */
			n += (speed / 2);
			eno = (unsigned int)(n / speed);
		
			if ( eno >= 8 ) {
				baud++;
				eno = 0;
			} 
		}
	}

	outb(val | UART_LCR_DLAB, info->base + UART_LCR);  /* set DLAB */
	outb((unsigned char)(baud & 0x00FF), info->base + UART_DLL);	    /* LS of divisor */
	outb((unsigned char)((baud >> 8) & 0xFF), info->base + UART_DLM); 	    /* MS of divisor */
SET_MOXA_MUST_ENUM_VALUE(info->base,(unsigned char)eno);
	outb(val, info->base + UART_LCR);		    /* reset DLAB */
}

# if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
module_init(mxpcdrv_module_init);
module_exit(mxpcdrv_module_exit); 
#endif

